# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Abstract linear vector spaces."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from builtins import object, range, str
from future import standard_library
standard_library.install_aliases()

from abc import abstractmethod
import numpy as np

from odl.set.sets import Field, Set, UniversalSet


__all__ = ('LinearSpace', 'LinearSpaceElement', 'UniversalSpace',
           'LinearSpaceTypeError', 'LinearSpaceTypeError')


class LinearSpace(Set):
    """Abstract linear vector space.

    Its elements are represented as instances of the
    `LinearSpaceElement` class.
    """

    def __init__(self, field):
        """Initialize a new instance.

        This method should be called by all inheriting methods so that
        the `field` property of the space is properly set.

        Parameters
        ----------
        field : `Field`
            Scalar field of numbers for this space.
        """
        if not isinstance(field, Field):
            raise TypeError('`field` {!r} is not a `Field` instance')
        self.__field = field

    @property
    def field(self):
        """Scalar field of numbers for this vector space."""
        return self.__field

    @abstractmethod
    def element(self, inp=None, **kwargs):
        """Create a `LinearSpaceElement` from ``inp`` or from scratch.

        If called without ``inp`` argument, an arbitrary element of the
        space is generated without guarantee of its state.

        If ``inp in self``, this has return ``inp`` or a view of ``inp``,
        otherwise, a copy may or may not occur.

        Parameters
        ----------
        inp : optional
            Input data from which to create the element.
        kwargs :
            Optional further arguments.

        Returns
        -------
        element : `LinearSpaceElement`
            A new element of this space.
        """

    @property
    def examples(self):
        """Example elements `zero` and `one` (if available)."""
        # All spaces should yield the zero element
        yield ('Zero', self.zero())

        try:
            yield ('One', self.one())
        except NotImplementedError:
            pass

    @abstractmethod
    def _lincomb(self, a, x1, b, x2, out):
        """Implement ``out[:] = a * x1 + b * x2``.

        This method is intended to be private. Public callers should
        resort to `lincomb` which is type-checked.
        """

    def _dist(self, x1, x2):
        """Return the distance between ``x1`` and ``x2``.

        This method is intended to be private. Public callers should
        resort to `dist` which is type-checked.
        """
        # default implementation
        return self.norm(x1 - x2)

    def _norm(self, x):
        """Return the norm of ``x``.

        This method is intended to be private. Public callers should
        resort to `norm` which is type-checked.
        """
        # default implementation
        return float(np.sqrt(self.inner(x, x).real))

    def _inner(self, x1, x2):
        """Return the inner product of ``x1`` and ``x2``.

        This method is intended to be private. Public callers should
        resort to `inner` which is type-checked.
        """
        # No default implementation possible
        raise LinearSpaceNotImplementedError(
            'inner product not implemented in space {!r}'.format(self))

    def _multiply(self, x1, x2, out):
        """Implement the pointwise multiplication ``out[:] = x1 * x2``.

        This method is intended to be private. Public callers should
        resort to `multiply` which is type-checked.
        """
        # No default implementation possible
        raise LinearSpaceNotImplementedError(
            'multiplication not implemented in space {!r}'.format(self))

    def one(self):
        """Return the one (multiplicative unit) element of this space."""
        raise LinearSpaceNotImplementedError(
            '`one` element not implemented in space {!r}'.format(self))

    # Default methods
    def zero(self):
        """Return the zero (additive unit) element of this space."""
        # Default implementation using lincomb
        tmp = self.element()
        self.lincomb(tmp, 0, tmp, 0, tmp)
        return tmp

    def __contains__(self, other):
        """Return ``other in self``.

        Returns
        -------
        contains : bool
            ``True`` if ``other`` is a `LinearSpaceElement` instance and
            ``other.space`` is equal to this space, ``False`` otherwise.

        Notes
        -----
        This is the strict default where spaces must be equal.
        Subclasses may choose to implement a less strict check.
        """
        return getattr(other, 'space', None) == self

    # Error checking variant of methods
    def lincomb(self, a, x1, b=None, x2=None, out=None):
        """Implement ``out[:] = a * x1 + b * x2``.

        This function implements

            ``out[:] = a * x1``

        or, if ``b`` and ``y`` are given,

            ``out = a * x1 + b * x2``.

        Parameters
        ----------
        a : `field` element
            Scalar to multiply ``x1`` with.
        x1 : `LinearSpaceElement`
            First space element in the linear combination.
        b : `field` element, optional
            Scalar to multiply ``x2`` with. Required if ``x2`` is
            provided.
        x2 : `LinearSpaceElement`, optional
            Second space element in the linear combination.
        out : `LinearSpaceElement`, optional
            Element to which the result is written.

        Returns
        -------
        out : `LinearSpaceElement`
            Result of the linear combination. If ``out`` was provided,
            the returned object is a reference to it.

        Notes
        -----
        The elements ``out``, ``x1`` and ``x2`` may be aligned, thus a call

            ``space.lincomb(x, 2, x, 3.14, out=x)``

        is (mathematically) equivalent to

            ``x = x * (1 + 2 + 3.14)``.
        """
        if out is None:
            out = self.element()

        if out not in self:
            raise LinearSpaceTypeError('`out` {!r} is not an element of {!r}'
                                       ''.format(out, self))
        if a not in self.field:
            raise LinearSpaceTypeError('`a` {!r} not an element of the field '
                                       '{!r} of {!r}'
                                       ''.format(a, self.field, self))
        if x1 not in self:
            raise LinearSpaceTypeError('`x1` {!r} is not an element of {!r}'
                                       ''.format(x1, self))

        if b is None:  # Single element
            if x2 is not None:
                raise ValueError('`x2` provided but not `b`')
            self._lincomb(a, x1, 0, x1, out)
            return out

        else:  # Two elements
            if b not in self.field:
                raise LinearSpaceTypeError('`b` {!r} not an element of the '
                                           'field {!r} of {!r}'
                                           ''.format(b, self.field, self))
            if x2 not in self:
                raise LinearSpaceTypeError('`x2` {!r} is not an element of '
                                           '{!r}'.format(x2, self))

            self._lincomb(a, x1, b, x2, out)

        return out

    def dist(self, x1, x2):
        """Return the distance between ``x1`` and ``x2``.

        Parameters
        ----------
        x1, x2 : `LinearSpaceElement`
            Elements whose distance to compute.

        Returns
        -------
        dist : float
            Distance between ``x1`` and ``x2``.
        """
        if x1 not in self:
            raise LinearSpaceTypeError('`x1` {!r} is not an element of '
                                       '{!r}'.format(x1, self))
        if x2 not in self:
            raise LinearSpaceTypeError('`x2` {!r} is not an element of '
                                       '{!r}'.format(x2, self))

        return float(self._dist(x1, x2))

    def norm(self, x):
        """Return the norm of ``x``.

        Parameters
        ----------
        x : `LinearSpaceElement`
            Element whose norm to compute.

        Returns
        -------
        norm : float
            Norm of ``x``.
        """
        if x not in self:
            raise LinearSpaceTypeError('`x` {!r} is not an element of '
                                       '{!r}'.format(x, self))

        return float(self._norm(x))

    def inner(self, x1, x2):
        """Return the inner product of ``x1`` and ``x2``.

        Parameters
        ----------
        x1, x2 : `LinearSpaceElement`
            Elements whose inner product to compute.

        Returns
        -------
        inner : `LinearSpace.field` element
            Inner product of ``x1`` and ``x2``.
        """
        if x1 not in self:
            raise LinearSpaceTypeError('`x1` {!r} is not an element of '
                                       '{!r}'.format(x1, self))
        if x2 not in self:
            raise LinearSpaceTypeError('`x2` {!r} is not an element of '
                                       '{!r}'.format(x2, self))

        return self.field.element(self._inner(x1, x2))

    def multiply(self, x1, x2, out=None):
        """Return the pointwise product of ``x1`` and ``x2``.

        Parameters
        ----------
        x1, x2 : `LinearSpaceElement`
            Multiplicands in the product.
        out : `LinearSpaceElement`, optional
            Element to which the result is written.

        Returns
        -------
        out : `LinearSpaceElement`
            Product of the elements. If ``out`` was provided, the
            returned object is a reference to it.
        """
        if out is None:
            out = self.element()

        if out not in self:
            raise LinearSpaceTypeError('`out` {!r} is not an element of '
                                       '{!r}'.format(out, self))
        if x1 not in self:
            raise LinearSpaceTypeError('`x1` {!r} is not an element of '
                                       '{!r}'.format(x1, self))
        if x2 not in self:
            raise LinearSpaceTypeError('`x2` {!r} is not an element of '
                                       '{!r}'.format(x2, self))

        self._multiply(x1, x2, out)
        return out

    def divide(self, x1, x2, out=None):
        """Return the pointwise quotient of ``x1`` and ``x2``

        Parameters
        ----------
        x1 : `LinearSpaceElement`
            Dividend in the quotient.
        x2 : `LinearSpaceElement`
            Divisor in the quotient.
        out : `LinearSpaceElement`, optional
            Element to which the result is written.

        Returns
        -------
        out : `LinearSpaceElement`
            Quotient of the elements. If ``out`` was provided, the
            returned object is a reference to it.
        """
        if out is None:
            out = self.element()

        if out not in self:
            raise LinearSpaceTypeError('`out` {!r} is not an element of '
                                       '{!r}'.format(out, self))
        if x1 not in self:
            raise LinearSpaceTypeError('`x1` {!r} is not an element of '
                                       '{!r}'.format(x1, self))
        if x2 not in self:
            raise LinearSpaceTypeError('`x2` {!r} is not an element of '
                                       '{!r}'.format(x2, self))

        self._divide(x1, x2, out)
        return out

    @property
    def element_type(self):
        """Type of elements of this space (`LinearSpaceElement`)."""
        return LinearSpaceElement


class LinearSpaceElement(object):

    """Abstract class for `LinearSpace` elements.

    Do not use this class directly -- to create an element of a vector
    space, call the space's `LinearSpace.element` method instead.
    """

    def __init__(self, space):
        """Initialize a new instance.

        All deriving classes must call this method to set the `space`
        property.
        """
        if not isinstance(space, LinearSpace):
            raise TypeError('`space` {!r} is not a `LinearSpace` instance'
                            ''.format(space))
        self.__space = space

    @property
    def space(self):
        """Space to which this element belongs."""
        return self.__space

    # Convenience functions
    def assign(self, other):
        """Assign the values of ``other`` to ``self``."""
        return self.space.lincomb(1, other, out=self)

    def copy(self):
        """Create an identical (deep) copy of self."""
        result = self.space.element()
        result.assign(self)
        return result

    def lincomb(self, a, x1, b=None, x2=None):
        """Implement ``self[:] = a * x1 + b * x2``.

        Parameters
        ----------
        a : element of ``space.field``
            Scalar to multiply ``x1`` with.
        x1 : `LinearSpaceElement`
            First space element in the linear combination.
        b : element of ``space.field``, optional
            Scalar to multiply ``x2`` with. Required if ``x2`` is
            provided.
        x2 : `LinearSpaceElement`, optional
            Second space element in the linear combination.

        See Also
        --------
        LinearSpace.lincomb
        """
        return self.space.lincomb(a, x1, b, x2, out=self)

    def set_zero(self):
        """Set this element to zero.

        See Also
        --------
        LinearSpace.zero
        """
        return self.space.lincomb(0, self, 0, self, out=self)

    # Convenience methods
    def __iadd__(self, other):
        """Implement ``self += other``."""
        if other in self.space:
            return self.space.lincomb(1, self, 1, other, out=self)
        elif other in self.space.field:
            one = getattr(self.space, 'one', None)
            if one is None:
                return NotImplemented
            else:
                # other --> other * space.one()
                return self.space.lincomb(1, self, other, one(), out=self)
        else:
            # We do not `return NotImplemented` here since we don't want a
            # fallback for in-place. Otherwise python attempts
            # `self = self + other` which does not modify self.
            raise TypeError('cannot add {!r} and {!r} in place'
                            ''.format(self, other))

    def __add__(self, other):
        """Return ``self + other``."""
        # Instead of using __iadd__ we duplicate code here for performance
        if other in self.space:
            tmp = self.space.element()
            return self.space.lincomb(1, self, 1, other, out=tmp)
        elif other in self.space.field:
            one = getattr(self.space, 'one', None)
            if one is None:
                return NotImplemented
            else:
                tmp = one()
                return self.space.lincomb(1, self, other, tmp, out=tmp)
        else:
            return NotImplemented

    __radd__ = __add__

    def __isub__(self, other):
        """Implement ``self -= other``."""
        if other in self.space:
            return self.space.lincomb(1, self, -1, other, out=self)
        elif other in self.space.field:
            one = getattr(self.space, 'one', None)
            if one is None:
                return NotImplemented
            else:
                return self.space.lincomb(1, self, -other, one(), out=self)
        else:
            # We do not `return NotImplemented` here since we don't want a
            # fallback for in-place. Otherwise python attempts
            # `self = self - other` which does not modify self.
            raise TypeError('cannot subtract {!r} and {!r} in place'
                            ''.format(self, other))

    def __sub__(self, other):
        """Return ``self - other``."""
        # Instead of using __isub__ we duplicate code here for performance
        if other in self.space:
            tmp = self.space.element()
            return self.space.lincomb(1, self, -1, other, out=tmp)
        elif other in self.space.field:
            one = getattr(self.space, 'one', None)
            if one is None:
                return NotImplemented
            else:
                tmp = one()
                return self.space.lincomb(1, self, -other, tmp, out=tmp)
        else:
            return NotImplemented

    def __rsub__(self, other):
        """Return ``other - self``."""
        if other in self.space:
            tmp = self.space.element()
            return self.space.lincomb(1, other, -1, self, out=tmp)
        if other in self.space.field:
            one = getattr(self.space, 'one', None)
            if one is None:
                return NotImplemented
            else:
                # other --> other * space.one()
                tmp = one()
                self.space.lincomb(other, tmp, out=tmp)
                return self.space.lincomb(1, tmp, -1, self, out=tmp)
        else:
            return NotImplemented

    def __imul__(self, other):
        """Implement ``self *= other``."""
        if other in self.space.field:
            return self.space.lincomb(other, self, out=self)
        elif other in self.space:
            return self.space.multiply(other, self, out=self)
        else:
            # We do not `return NotImplemented` here since we don't want a
            # fallback for in-place. Otherwise python attempts
            # `self = self * other` which does not modify self.
            raise TypeError('cannot multiply {!r} and {!r} in place'
                            ''.format(self, other))

    def __mul__(self, other):
        """Return ``self * other``."""
        # Instead of using __imul__ we duplicate code here for performance
        if other in self.space.field:
            tmp = self.space.element()
            return self.space.lincomb(other, self, out=tmp)
        elif other in self.space:
            tmp = self.space.element()
            return self.space.multiply(other, self, out=tmp)
        else:
            return NotImplemented

    __rmul__ = __mul__

    def __itruediv__(self, other):
        """Implement ``self /= other``."""
        if other in self.space.field:
            return self.space.lincomb(1.0 / other, self, out=self)
        elif other in self.space:
            return self.space.divide(self, other, out=self)
        else:
            # We do not `return NotImplemented` here since we don't want a
            # fallback for in-place. Otherwise python attempts
            # `self = self / other` which does not modify self.
            raise TypeError('cannot divide {!r} and {!r} in place'
                            ''.format(self, other))

    __idiv__ = __itruediv__

    def __truediv__(self, other):
        """Return ``self / other``."""
        if other in self.space.field:
            tmp = self.space.element()
            return self.space.lincomb(1.0 / other, self, out=tmp)
        elif other in self.space:
            tmp = self.space.element()
            return self.space.divide(self, other, out=tmp)
        else:
            return NotImplemented

    __div__ = __truediv__

    def __rtruediv__(self, other):
        """Return ``other / self``."""
        if other in self.space.field:
            one = getattr(self.space, 'one', None)
            if one is None:
                return NotImplemented
            else:
                # other --> other * space.one()
                tmp = one()
                self.space.lincomb(other, tmp, out=tmp)
                return self.space.divide(tmp, self, out=tmp)
        elif other in self.space:
            tmp = self.space.element()
            return self.space.divide(other, self, out=tmp)
        else:
            return NotImplemented

    __rdiv__ = __rtruediv__

    def __ipow__(self, p):
        """Implement ``self ** p``.

        This is only defined for integer ``p``."""
        p, p_in = int(p), p
        if p != p_in:
            raise ValueError('expected integer `p`, got {}'.format(p_in))
        if p < 0:
            self **= -p
            self.space.divide(self.space.one(), self, out=self)
            return self
        elif p == 0:
            self.assign(self.space.one())
            return self
        elif p == 1:
            return self
        elif p % 2 == 0:
            self.space.multiply(self, self, out=self)
            self **= p // 2
            return self
        else:
            tmp = self.copy()
            for _ in range(p - 2):
                self.space.multiply(tmp, self, out=tmp)
            self.space.multiply(tmp, self, out=self)
            return self

    def __pow__(self, p):
        """Return ``self ** p``."""
        tmp = self.copy()
        tmp.__ipow__(p)
        return tmp

    def __neg__(self):
        """Return ``-self``."""
        return (-1) * self

    def __pos__(self):
        """Return ``+self``."""
        return self.copy()

    # Metric space method
    def __eq__(self, other):
        """Return ``self == other``.

        Two elements are equal if their distance is zero.

        Parameters
        ----------
        other : `LinearSpaceElement`
            Element of this space.

        Returns
        -------
        equals : bool
            ``True`` if the elements are equal ``False`` otherwise.

        See Also
        --------
        LinearSpace.dist

        Notes
        -----
        Equality is very sensitive to numerical errors, thus any
        arithmetic operations should be expected to break equality.

        Examples
        --------
        >>> rn = odl.rn(1, norm=np.linalg.norm)
        >>> x = rn.element([0.1])
        >>> x == x
        True
        >>> y = rn.element([0.1])
        >>> x == y
        True
        >>> z = rn.element([0.3])
        >>> x + x + x == z
        False
        """
        if other is self:
            # Optimization for a common case
            return True
        elif (not isinstance(other, LinearSpaceElement) or
              other.space != self.space):
            # Cannot use (if other not in self.space) since this is not
            # reflexive.
            return False
        else:
            return self.space.dist(self, other) == 0

    def __ne__(self, other):
        """Return ``self != other``."""
        return not self.__eq__(other)

    def __str__(self):
        """Return ``str(self)``.

        This is a default implementation that does not print any
        information about the contents of the element.
        """
        return str(self.space) + 'Element'

    def __repr__(self):
        """Return ``repr(self)``.

        This is a default implementation that does not print any
        information about the contents of the element.
        """
        return repr(self.space) + 'Element'

    def __copy__(self):
        """Return a copy of this element.

        See Also
        --------
        LinearSpace.copy
        """
        return self.copy()

    def __deepcopy__(self, memo):
        """Return a deep copy of this element.

        See Also
        --------
        LinearSpace.copy
        """
        return self.copy()

    def norm(self):
        """Return the norm of this element.

        See Also
        --------
        LinearSpace.norm
        """
        return self.space.norm(self)

    def dist(self, other):
        """Return the distance of ``self`` to ``other``.

        See Also
        --------
        LinearSpace.dist
        """
        return self.space.dist(self, other)

    def inner(self, other):
        """Return the inner product of ``self`` and ``other``.

        See Also
        --------
        LinearSpace.inner
        """
        return self.space.inner(self, other)

    def multiply(self, other, out=None):
        """Return ``out = self * other``.

        If ``out`` is provided, the result is written to it.

        See Also
        --------
        LinearSpace.multiply
        """
        return self.space.multiply(self, other, out=out)

    def divide(self, other, out=None):
        """Return ``out = self / other``.

        If ``out`` is provided, the result is written to it.

        See Also
        --------
        LinearSpace.divide
        """
        return self.space.divide(self, other, out=out)

    @property
    def T(self):
        """This element's transpose, i.e. the functional ``<. , self>``.

        Returns
        -------
        transpose : `InnerProductOperator`

        Notes
        -----
        This function is only defined in inner product spaces.

        In a complex space, the conjugate transpose of is taken instead
        of the transpose only.

        Examples
        --------
        >>> rn = odl.rn(3)
        >>> x = rn.element([1, 2, 3])
        >>> y = rn.element([2, 1, 3])
        >>> x.T(y)
        13.0
        """
        from odl.operator import InnerProductOperator
        return InnerProductOperator(self.copy())

    # Give an `Element` a higher priority than any NumPy array type. This
    # forces the usage of `__op__` of `Element` if the other operand
    # is a NumPy object (applies also to scalars!).
    __array_priority__ = 1000000.0


class UniversalSpace(LinearSpace):

    """A dummy linear space class.

    Mostly raising `LinearSpaceNotImplementedError`.
    """

    def __init__(self):
        """Initialize a new instance."""
        LinearSpace.__init__(self, field=UniversalSet())

    def element(self, inp=None):
        """Dummy element creation method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _lincomb(self, a, x1, b, x2, out):
        """Dummy linear combination.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _dist(self, x1, x2):
        """Dummy distance method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _norm(self, x):
        """Dummy norm method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _inner(self, x1, x2):
        """Dummy inner product method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def _multiply(self, x1, x2, out):
        """Dummy multiplication method.

        raises `LinearSpaceNotImplementedError`."""
        raise LinearSpaceNotImplementedError

    def _divide(self, x1, x2, out):
        """Dummy division method.

        raises `LinearSpaceNotImplementedError`.
        """
        raise LinearSpaceNotImplementedError

    def __eq__(self, other):
        """Return ``self == other``.

        Dummy check, ``True`` for any `LinearSpace`.
        """
        return isinstance(other, LinearSpace)

    def __contains__(self, other):
        """Return ``other in self``.

        Dummy membership check, ``True`` for any `LinearSpaceElement`.
        """
        return isinstance(other, LinearSpaceElement)


class LinearSpaceTypeError(TypeError):
    """Exception for type errors in `LinearSpace`'s.

    This exception is raised when the wrong type of element is fed to
    `LinearSpace.lincomb` and related functions.
    """


class LinearSpaceNotImplementedError(NotImplementedError):
    """Exception for unimplemented functionality in `LinearSpace`'s.

    This exception is raised when a method is called in `LinearSpace`
    that has not been defined in a specific space.
    """


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.


"""Back-end for STIR: Software for Tomographic Reconstruction.

Back and forward projectors for PET.

`ForwardProjectorByBinWrapper` and `BackProjectorByBinWrapper` are general
objects of STIR projectors and back-projectors, these can be used to wrap a
given projector.

`stir_projector_from_file` allows users a easy way to create a
`ForwardProjectorByBinWrapper` by giving file paths to the required templates.

References
----------
See the `STIR webpage`_ for more information and the `STIR doc`_ for info on
the STIR classes used here.

.. _STIR webpage: http://stir.sourceforge.net
.. _STIR doc: http://stir.sourceforge.net/documentation/doxy/html/
"""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

try:
    import stir
    # Fix for stirextra being moved around in various stir versions
    try:
        stirextra = stir.stirextra
    except AttributeError:
        import stirextra

    STIR_AVAILABLE = True
except ImportError:
    STIR_AVAILABLE = False

from odl.discr import uniform_discr
from odl.operator import Operator


__all__ = ('ForwardProjectorByBinWrapper',
           'BackProjectorByBinWrapper',
           'stir_projector_from_file',
           'STIR_AVAILABLE')


class StirVerbosity(object):

    """Context manager setting STIR verbosity to a fixed level."""

    def __init__(self, verbosity):
        self.verbosity = verbosity
        self.old_verbosity = None

    def __enter__(self):
        self.old_verbosity = stir.Verbosity.get()
        stir.Verbosity.set(self.verbosity)

    def __exit__(self, *_):
        stir.Verbosity.set(self.old_verbosity)


class ForwardProjectorByBinWrapper(Operator):

    """A forward projector using STIR.

    Uses "ForwardProjectorByBinUsingProjMatrixByBin" as a projector.
    """

    def __init__(self, domain, range, volume, proj_data,
                 projector=None, adjoint=None):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `DiscreteLp`
            Volume of the projection. Needs to have the same shape as
            ``volume.shape()``.
        range : `DiscreteLp`
            Projection space. Needs to have the same shape as
            ``proj_data.to_array().shape()``.
        volume : ``stir.FloatVoxelsOnCartesianGrid``
            Stir volume to use in the forward projection
        proj_data : ``stir.ProjData``
            Stir description of the projection.
        projector : ``stir.ForwardProjectorByBin``, optional
            A pre-initialized projector.
        adjoint : `BackProjectorByBinWrapper`, optional
            A pre-initialized adjoint.
        """
        # Check data sizes
        if domain.shape != volume.shape():
            raise ValueError('domain.shape {} does not equal volume shape {}'
                             ''.format(domain.shape, volume.shape()))
        # TODO: improve
        proj_shape = proj_data.to_array().shape()
        if range.shape != proj_shape:
            raise ValueError('range.shape {} does not equal proj shape {}'
                             ''.format(range.shape, proj_shape))

        # Set domain, range etc
        super().__init__(domain, range, True)

        # Read template of the projection
        self.proj_data = proj_data
        self.proj_data_info = proj_data.get_proj_data_info()
        self.volume = volume

        # Create forward projection by matrix
        if projector is None:
            proj_matrix = stir.ProjMatrixByBinUsingRayTracing()
            proj_matrix.set_up(self.proj_data_info, self.volume)
            self.projector = stir.ForwardProjectorByBinUsingProjMatrixByBin(
                proj_matrix)
            self.projector.set_up(self.proj_data_info, self.volume)

            # If no adjoint was given, we initialize a projector here to
            # save time.
            if adjoint is None:
                back_projector = stir.BackProjectorByBinUsingProjMatrixByBin(
                    proj_matrix)
                back_projector.set_up(self.proj_data.get_proj_data_info(),
                                      self.volume)
        else:
            # If user wants to provide both a projector and a back-projector,
            # he should wrap the back projector in an Operator
            self.projector = projector
            back_projector = None

        # Pre-create an adjoint to save time
        if adjoint is None:
            self._adjoint = BackProjectorByBinWrapper(
                self.range, self.domain, self.volume, self.proj_data,
                back_projector, self)
        else:
            self._adjoint = adjoint

    def _call(self, volume, out):
        """Forward project a volume."""
        # Set volume data
        self.volume.fill(volume.asarray().flat)

        # project
        with StirVerbosity(0):
            self.projector.forward_project(self.proj_data, self.volume)

        # make ODL data
        out[:] = stirextra.to_numpy(self.proj_data)

    @property
    def adjoint(self):
        """Back-projector associated with this operator."""
        return self._adjoint


class BackProjectorByBinWrapper(Operator):

    """A back projector using STIR."""

    def __init__(self, domain, range, volume, proj_data,
                 back_projector=None, adjoint=None):
        """Initialize a new instance.

        Parameters
        ----------
        domain : `DiscreteLp`
            Projection space. Needs to have the same shape as
            ``proj_data.to_array().shape()``.
        range : `DiscreteLp`
            Volume of the projection. Needs to have the same shape as
            ``volume.shape()``.
        volume : ``stir.FloatVoxelsOnCartesianGrid``
            Stir volume to use in the forward projection
        proj_data : ``stir.ProjData``
            Stir description of the projection.
        back_projector : ``stir.BackProjectorByBin``, optional
            A pre-initialized back-projector.
        adjoint : `ForwardProjectorByBinWrapper`, optional
            A pre-initialized adjoint.

        Notes
        -----
        See `STIR doc`_ for info on the STIR classes.

        References
        ----------
        .. _STIR doc: http://stir.sourceforge.net/documentation/doxy/html/
        """

        # Check data sizes
        if range.shape != volume.shape():
            raise ValueError('`range.shape` {} does not equal volume shape {}'
                             ''.format(range.shape, volume.shape()))
        # TODO: improve
        proj_shape = proj_data.to_array().shape()
        if domain.shape != proj_shape:
            raise ValueError('`domain.shape` {} does not equal proj shape {}'
                             ''.format(range.shape, proj_shape))

        # Set range domain
        super().__init__(domain, range, True)

        # Read template of the projection
        self.proj_data = proj_data
        self.proj_data_info = proj_data.get_proj_data_info()
        self.volume = volume

        # Create forward projection by matrix
        if back_projector is None:
            proj_matrix = stir.ProjMatrixByBinUsingRayTracing()
            proj_matrix.set_up(self.proj_data_info, self.volume)

            self.back_projector = stir.BackProjectorByBinUsingProjMatrixByBin(
                proj_matrix)
            self.back_projector.set_up(self.proj_data.get_proj_data_info(),
                                       self.volume)

            if adjoint is None:
                projector = stir.ForwardProjectorByBinUsingProjMatrixByBin(
                    proj_matrix)
                projector.set_up(self.proj_data_info, self.volume)

        else:
            self.back_projector = back_projector
            projector = None

        # Pre-create an adjoint to save time
        if adjoint is None:
            self._adjoint = ForwardProjectorByBinWrapper(
                self.range, self.domain, self.volume, self.proj_data,
                projector, self)
        else:
            self._adjoint = adjoint

    def _call(self, projections, out):
        """Back project."""
        # Set projection data
        self.proj_data.fill(projections.asarray().flat)

        # back-project
        with StirVerbosity(0):
            self.back_projector.back_project(self.volume, self.proj_data)

        # make ODL data
        out[:] = stirextra.to_numpy(self.volume)


def stir_projector_from_file(volume_file, projection_file):
    """Create a STIR projector from given template files.

    Parameters
    ----------
    volume_file : string
        Full file path to the STIR input file containing information on the
        volume. This is usually a '.hv' file. For STIR reasons,
        a '.v' file is also needed.
    projection_file : string
        Full file path to the STIR input file with information on the
        projection data. This is usually a '.hs' file. For STIR reasons,
        a '.s' file is also needed.

    Returns
    -------
    projector : `ForwardProjectorByBinWrapper`
        A STIR forward projector.
    """
    volume = stir.FloatVoxelsOnCartesianGrid.read_from_file(volume_file)

    proj_data_in = stir.ProjData.read_from_file(projection_file)
    proj_data = stir.ProjDataInMemory(proj_data_in.get_exam_info(),
                                      proj_data_in.get_proj_data_info())

    origin = volume.get_origin()
    grid_spacing = volume.get_grid_spacing()
    grid_shape = [volume.get_z_size(),
                  volume.get_y_size(),
                  volume.get_x_size()]
    min_pt = [origin[1], origin[2], origin[3]]
    max_pt = [origin[1] + grid_spacing[1] * grid_shape[0],
              origin[2] + grid_spacing[2] * grid_shape[1],
              origin[3] + grid_spacing[3] * grid_shape[2]]

    # reverse to handle STIR bug? See:
    # https://github.com/UCL/STIR/issues/7
    recon_sp = uniform_discr(min_pt, max_pt, grid_shape,
                             dtype='float32')

    # TODO: set correct projection space. Currently, a default grid with
    # stride (1, 1, 1) is used.
    proj_shape = proj_data.to_array().shape()
    data_sp = uniform_discr([0, 0, 0], proj_shape, proj_shape, dtype='float32')

    return ForwardProjectorByBinWrapper(recon_sp, data_sp, volume, proj_data)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Numerical helper functions for convenience or speed."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

import numpy as np

from odl.util.normalize import normalized_scalar_param_list, safe_int_conv


__all__ = ('apply_on_boundary', 'fast_1d_tensor_mult', 'resize_array')


_SUPPORTED_RESIZE_PAD_MODES = ('constant', 'symmetric', 'periodic',
                               'order0', 'order1')


def apply_on_boundary(array, func, only_once=True, which_boundaries=None,
                      axis_order=None, out=None):
    """Apply a function of the boundary of an n-dimensional array.

    All other values are preserved as is.

    Parameters
    ----------
    array : `array-like`
        Modify the boundary of this array
    func : callable or sequence of callables
        If a single function is given, assign
        ``array[slice] = func(array[slice])`` on the boundary slices,
        e.g. use ``lamda x: x / 2`` to divide values by 2.
        A sequence of functions is applied per axis separately. It
        must have length ``array.ndim`` and may consist of one function
        or a 2-tuple of functions per axis.
        ``None`` entries in a sequence cause the axis (side) to be
        skipped.
    only_once : bool, optional
        If ``True``, ensure that each boundary point appears in exactly
        one slice. If ``func`` is a list of functions, the
        ``axis_order`` determines which functions are applied to nodes
        which appear in multiple slices, according to the principle
        "first-come, first-served".
    which_boundaries : sequence, optional
        If provided, this sequence determines per axis whether to
        apply the function at the boundaries in each axis. The entry
        in each axis may consist in a single bool or a 2-tuple of
        bool. In the latter case, the first tuple entry decides for
        the left, the second for the right boundary. The length of the
        sequence must be ``array.ndim``. ``None`` is interpreted as
        "all boundaries".
    axis_order : sequence of ints, optional
        Permutation of ``range(array.ndim)`` defining the order in which
        to process the axes. If combined with ``only_once`` and a
        function list, this determines which function is evaluated in
        the points that are potentially processed multiple times.
    out : `numpy.ndarray`, optional
        Location in which to store the result, can be the same as ``array``.
        Default: copy of ``array``

    Examples
    --------
    >>> arr = np.ones((3, 3))
    >>> apply_on_boundary(arr, lambda x: x / 2)
    array([[ 0.5,  0.5,  0.5],
           [ 0.5,  1. ,  0.5],
           [ 0.5,  0.5,  0.5]])

    If called with ``only_once=False``, the function is applied repeatedly:

    >>> apply_on_boundary(arr, lambda x: x / 2, only_once=False)
    array([[ 0.25,  0.5 ,  0.25],
           [ 0.5 ,  1.  ,  0.5 ],
           [ 0.25,  0.5 ,  0.25]])

    >>> apply_on_boundary(arr, lambda x: x / 2, only_once=True,
    ...                   which_boundaries=((True, False), True))
    array([[ 0.5,  0.5,  0.5],
           [ 0.5,  1. ,  0.5],
           [ 0.5,  1. ,  0.5]])

    Use the ``out`` parameter to store the result in an existing array:

    >>> out = np.empty_like(arr)
    >>> result = apply_on_boundary(arr, lambda x: x / 2, out=out)
    >>> result
    array([[ 0.5,  0.5,  0.5],
           [ 0.5,  1. ,  0.5],
           [ 0.5,  0.5,  0.5]])
    >>> result is out
    True
    """
    array = np.asarray(array)

    if callable(func):
        func = [func] * array.ndim
    elif len(func) != array.ndim:
        raise ValueError('sequence of functions has length {}, expected {}'
                         ''.format(len(func), array.ndim))

    if which_boundaries is None:
        which_boundaries = ([(True, True)] * array.ndim)
    elif len(which_boundaries) != array.ndim:
        raise ValueError('`which_boundaries` has length {}, expected {}'
                         ''.format(len(which_boundaries), array.ndim))

    if axis_order is None:
        axis_order = list(range(array.ndim))
    elif len(axis_order) != array.ndim:
        raise ValueError('`axis_order` has length {}, expected {}'
                         ''.format(len(axis_order), array.ndim))

    if out is None:
        out = array.copy()
    else:
        out[:] = array  # Self assignment is free, in case out is array

    # The 'only_once' functionality is implemented by storing for each axis
    # if the left and right boundaries have been processed. This information
    # is stored in a list of slices which is reused for the next axis in the
    # list.
    slices = [slice(None)] * array.ndim
    for ax, function, which in zip(axis_order, func, which_boundaries):
        if only_once:
            slc_l = list(slices)  # Make a copy; copy() exists in Py3 only
            slc_r = list(slices)
        else:
            slc_l = [slice(None)] * array.ndim
            slc_r = [slice(None)] * array.ndim

        # slc_l and slc_r select left and right boundary, resp, in this axis.
        slc_l[ax] = 0
        slc_r[ax] = -1

        try:
            # Tuple of functions in this axis
            func_l, func_r = function
        except TypeError:
            # Single function
            func_l = func_r = function

        try:
            # Tuple of bool
            mod_left, mod_right = which
        except TypeError:
            # Single bool
            mod_left = mod_right = which

        if mod_left and func_l is not None:
            out[slc_l] = func_l(out[slc_l])
            start = 1
        else:
            start = None

        if mod_right and func_r is not None:
            out[slc_r] = func_r(out[slc_r])
            end = -1
        else:
            end = None

        # Write the information for the processed axis into the slice list.
        # Start and end include the boundary if it was processed.
        slices[ax] = slice(start, end)

    return out


def fast_1d_tensor_mult(ndarr, onedim_arrs, axes=None, out=None):
    """Fast multiplication of an n-dim array with an outer product.

    This method implements the multiplication of an n-dimensional array
    with an outer product of one-dimensional arrays, e.g.::

        a = np.ones((10, 10, 10))
        x = np.random.rand(10)
        a *= x[:, None, None] * x[None, :, None] * x[None, None, :]

    Basically, there are two ways to do such an operation:

    1. First calculate the factor on the right-hand side and do one
       "big" multiplication; or
    2. Multiply by one factor at a time.

    The procedure of building up the large factor in the first method
    is relatively cheap if the number of 1d arrays is smaller than the
    number of dimensions. For exactly n vectors, the second method is
    faster, although it loops of the array ``a`` n times.

    This implementation combines the two ideas into a hybrid scheme:

    - If there are less 1d arrays than dimensions, choose 1.
    - Otherwise, calculate the factor array for n-1 arrays
      and multiply it to the large array. Finally, multiply with the
      last 1d array.

    The advantage of this approach is that it is memory-friendly and
    loops over the big array only twice.

    Parameters
    ----------
    ndarr : `array-like`
        Array to multiply to
    onedim_arrs : sequence of `array-like`'s
        One-dimensional arrays to be multiplied with ``ndarr``. The
        sequence may not be longer than ``ndarr.ndim``.
    axes : sequence of ints, optional
        Take the 1d transform along these axes. ``None`` corresponds to
        the last ``len(onedim_arrs)`` axes, in ascending order.
    out : `numpy.ndarray`, optional
        Array in which the result is stored

    Returns
    -------
    out : `numpy.ndarray`
        Result of the modification. If ``out`` was given, the returned
        object is a reference to it.
    """
    if out is None:
        out = np.array(ndarr, copy=True)
    else:
        out[:] = ndarr  # Self-assignment is free if out is ndarr

    if not onedim_arrs:
        raise ValueError('no 1d arrays given')

    if axes is None:
        axes = list(range(out.ndim - len(onedim_arrs), out.ndim))
        axes_in = None
    elif len(axes) != len(onedim_arrs):
        raise ValueError('there are {} 1d arrays, but {} axes entries'
                         ''.format(len(onedim_arrs), len(axes)))
    else:
        # Make axes positive
        axes, axes_in = np.array(axes, dtype=int), axes
        axes[axes < 0] += out.ndim
        axes = list(axes)

    if np.any(np.array(axes) >= out.ndim) or np.any(np.array(axes) < 0):
        raise ValueError('`axes` {} out of bounds for {} dimensions'
                         ''.format(axes_in, out.ndim))

    # Make scalars 1d arrays and squeezable arrays 1d
    alist = [np.atleast_1d(np.asarray(a).squeeze()) for a in onedim_arrs]
    if any(a.ndim != 1 for a in alist):
        raise ValueError('only 1d arrays allowed')

    if len(axes) < out.ndim:
        # Make big factor array (start with 0d)
        factor = np.array(1.0)
        for ax, arr in zip(axes, alist):
            # Meshgrid-style slice
            slc = [None] * out.ndim
            slc[ax] = slice(None)
            factor = factor * arr[slc]

        out *= factor

    else:
        # Hybrid approach

        # Get the axis to spare for the final multiplication, the one
        # with the largest stride.
        axis_order = np.argsort(out.strides)
        last_ax = axis_order[-1]
        last_arr = alist[axes.index(last_ax)]

        # Build the semi-big array and multiply
        factor = np.array(1.0)
        for ax, arr in zip(axes, alist):
            if ax == last_ax:
                continue

            slc = [None] * out.ndim
            slc[ax] = np.s_[:]
            factor = factor * arr[slc]

        out *= factor

        # Finally multiply by the remaining 1d array
        slc = [None] * out.ndim
        slc[last_ax] = np.s_[:]
        out *= last_arr[slc]

    return out


def resize_array(arr, newshp, offset=None, pad_mode='constant', pad_const=0,
                 direction='forward', out=None):
    """Return the resized version of ``arr`` with shape ``newshp``.

    In axes where ``newshp > arr.shape``, padding is applied according
    to the supplied options.
    Where ``newshp < arr.shape``, the array is cropped to the new
    size.

    See `the online documentation
    <https://odlgroup.github.io/odl/math/resizing_ops.html>`_
    on resizing operators for mathematical details.

    Parameters
    ----------
    arr : `array-like`
        Array to be resized.
    newshp : sequence of ints
        Desired shape of the output array.
    offset: sequence of ints, optional
        Specifies how many entries are added to/removed from the "left"
        side (corresponding to low indices) of ``arr``.
    pad_mode : string, optional
        Method to be used to fill in missing values in an enlarged array.

        ``'constant'``: Fill with ``pad_const``.

        ``'symmetric'``: Reflect at the boundaries, not doubling the
        outmost values. This requires left and right padding sizes
        to be strictly smaller than the original array shape.

        ``'periodic'``: Fill in values from the other side, keeping
        the order. This requires left and right padding sizes to be
        at most as large as the original array shape.

        ``'order0'``: Extend constantly with the outmost values
        (ensures continuity).

        ``'order1'``: Extend with constant slope (ensures continuity of
        the first derivative). This requires at least 2 values along
        each axis where padding is applied.

    pad_const : scalar, optional
        Value to be used in the ``'constant'`` padding mode.
    direction : {'forward', 'adjoint'}
        Determines which variant of the resizing is applied.

        'forward' : in axes where ``out`` is larger than ``arr``,
        apply padding. Otherwise, restrict to the smaller size.

        'adjoint' : in axes where ``out`` is larger than ``arr``,
        apply zero-padding. Otherwise, restrict to the smaller size
        and add the outside contributions according to ``pad_mode``.

    out : `numpy.ndarray`, optional
        Array to write the result to. Must have shape ``newshp`` and
        be able to hold the data type of the input array.

    Returns
    -------
    resized : `numpy.ndarray`
        Resized array created according to the above rules. If ``out``
        was given, the returned object is a reference to it.

    Examples
    --------
    The input can be shrunk by simply providing a smaller size.
    By default, values are removed from the right. When enlarging,
    zero-padding is applied by default, and the zeros are added to
    the right side. That behavior can be changed with the ``offset``
    parameter:

    >>> from odl.util.numerics import resize_array
    >>> resize_array([1, 2, 3], (1,))
    array([1])
    >>> resize_array([1, 2, 3], (1,), offset=2)
    array([3])
    >>> resize_array([1, 2, 3], (6,))
    array([1, 2, 3, 0, 0, 0])
    >>> resize_array([1, 2, 3], (7,), offset=2)
    array([0, 0, 1, 2, 3, 0, 0])

    The padding constant can be changed, as well as the padding
    mode:

    >>> resize_array([1, 2, 3], (7,), pad_const=-1, offset=2)
    array([-1, -1,  1,  2,  3, -1, -1])
    >>> resize_array([1, 2, 3], (7,), pad_mode='periodic', offset=2)
    array([2, 3, 1, 2, 3, 1, 2])
    >>> resize_array([1, 2, 3], (7,), pad_mode='symmetric', offset=2)
    array([3, 2, 1, 2, 3, 2, 1])
    >>> resize_array([1, 2, 3], (7,), pad_mode='order0', offset=2)
    array([1, 1, 1, 2, 3, 3, 3])
    >>> resize_array([1, 2, 3], (7,), pad_mode='order1', offset=2)
    array([-1,  0,  1,  2,  3,  4,  5])

    Everything works for arbitrary number of dimensions:

    >>> # Take the middle two columns and extend rows symmetrically
    >>> resize_array([[1, 2, 3, 4],
    ...               [5, 6, 7, 8],
    ...               [9, 10, 11, 12]],
    ...               (5, 2), pad_mode='symmetric', offset=[1, 1])
    array([[ 6,  7],
           [ 2,  3],
           [ 6,  7],
           [10, 11],
           [ 6,  7]])
    >>> # Take the rightmost two columns and extend rows symmetrically
    >>> # downwards
    >>> resize_array([[1, 2, 3, 4],
    ...               [5, 6, 7, 8],
    ...               [9, 10, 11, 12]], (5, 2), pad_mode='symmetric',
    ...              offset=[0, 2])
    array([[ 3,  4],
           [ 7,  8],
           [11, 12],
           [ 7,  8],
           [ 3,  4]])
    """
    # Handle arrays and shapes
    try:
        newshp = tuple(newshp)
    except TypeError:
        raise TypeError('`newshp` must be a sequence, got {!r}'.format(newshp))

    if out is not None:
        if not isinstance(out, np.ndarray):
            raise TypeError('`out` must be a `numpy.ndarray` instance, got '
                            '{!r}'.format(out))
        if out.shape != newshp:
            raise ValueError('`out` must have shape {}, got {}'
                             ''.format(newshp, out.shape))

        order = 'C' if out.flags.c_contiguous else 'F'
        arr = np.asarray(arr, dtype=out.dtype, order=order)
        if arr.ndim != out.ndim:
            raise ValueError('number of axes of `arr` and `out` do not match '
                             '({} != {})'.format(arr.ndim, out.ndim))
    else:
        arr = np.asarray(arr)
        order = 'C' if arr.flags.c_contiguous else 'F'
        out = np.empty(newshp, dtype=arr.dtype, order=order)
        if len(newshp) != arr.ndim:
            raise ValueError('number of axes of `arr` and `len(newshp)` do '
                             'not match ({} != {})'
                             ''.format(arr.ndim, len(newshp)))

    # Handle offset
    if offset is None:
        offset = [0] * out.ndim
    else:
        offset = normalized_scalar_param_list(
            offset, out.ndim, param_conv=safe_int_conv, keep_none=False)

    # Handle padding
    pad_mode, pad_mode_in = str(pad_mode).lower(), pad_mode
    if pad_mode not in _SUPPORTED_RESIZE_PAD_MODES:
        raise ValueError("`pad_mode` '{}' not understood".format(pad_mode_in))

    if (pad_mode == 'constant' and
        not np.can_cast(pad_const, out.dtype) and
        any(n_new > n_orig
            for n_orig, n_new in zip(arr.shape, out.shape))):
        raise ValueError('`pad_const` {} cannot be safely cast to the data '
                         'type {} of the output array'
                         ''.format(pad_const, out.dtype))

    # Handle direction
    direction, direction_in = str(direction).lower(), direction
    if direction not in ('forward', 'adjoint'):
        raise ValueError("`direction` '{}' not understood"
                         "".format(direction_in))

    if direction == 'adjoint' and pad_mode == 'constant' and pad_const != 0:
        raise ValueError("`pad_const` must be 0 for 'adjoint' direction, "
                         "got {}".format(pad_const))

    if direction == 'forward' and pad_mode == 'constant' and pad_const != 0:
        out.fill(pad_const)
    else:
        out.fill(0)

    # Perform the resizing
    if direction == 'forward':
        if pad_mode == 'constant':
            # Constant padding does not require the helper function
            _assign_intersection(out, arr, offset)
        else:
            # First copy the inner part and use it for padding
            _assign_intersection(out, arr, offset)
            _apply_padding(out, arr, offset, pad_mode, 'forward')
    else:
        if pad_mode == 'constant':
            # Skip the padding helper
            _assign_intersection(out, arr, offset)
        else:
            # Apply adjoint padding to a copy of the input and copy the inner
            # part when finished
            tmp = arr.copy()
            _apply_padding(tmp, out, offset, pad_mode, 'adjoint')
            _assign_intersection(out, tmp, offset)

    return out


def _intersection_slice_tuples(lhs_arr, rhs_arr, offset):
    """Return tuples to yield the intersecting part of both given arrays.

    The returned slices ``lhs_slc`` and ``rhs_slc`` are such that
    ``lhs_arr[lhs_slc]`` and ``rhs_arr[rhs_slc]`` have the same shape.
    The ``offset`` parameter determines how much is skipped/added on the
    "left" side (small indices).
    """
    lhs_slc, rhs_slc = [], []
    for istart, n_lhs, n_rhs in zip(offset, lhs_arr.shape, rhs_arr.shape):

        # Slice for the inner part in the larger array corresponding to the
        # small one, offset by the given amount
        istop = istart + min(n_lhs, n_rhs)
        inner_slc = slice(istart, istop)

        if n_lhs > n_rhs:
            # Extension
            lhs_slc.append(inner_slc)
            rhs_slc.append(slice(None))
        elif n_lhs < n_rhs:
            # Restriction
            lhs_slc.append(slice(None))
            rhs_slc.append(inner_slc)
        else:
            # Same size, so full slices for both
            lhs_slc.append(slice(None))
            rhs_slc.append(slice(None))

    return tuple(lhs_slc), tuple(rhs_slc)


def _assign_intersection(lhs_arr, rhs_arr, offset):
    """Assign the intersecting region from ``rhs_arr`` to ``lhs_arr``."""
    lhs_slc, rhs_slc = _intersection_slice_tuples(lhs_arr, rhs_arr, offset)
    lhs_arr[lhs_slc] = rhs_arr[rhs_slc]


def _padding_slices_outer(lhs_arr, rhs_arr, axis, offset):
    """Return slices into the outer array part where padding is applied.

    When padding is performed, these slices yield the outer (excess) part
    of the larger array that is to be filled with values. Slices for
    both sides of the arrays in a given ``axis`` are returned.

    The same slices are used also in the adjoint padding correction,
    however in a different way.

    See `the online documentation
    <https://odlgroup.github.io/odl/math/resizing_ops.html>`_
    on resizing operators for details.
    """
    istart_inner = offset[axis]
    istop_inner = istart_inner + min(lhs_arr.shape[axis], rhs_arr.shape[axis])
    return slice(istart_inner), slice(istop_inner, None)


def _padding_slices_inner(lhs_arr, rhs_arr, axis, offset, pad_mode):
    """Return slices into the inner array part for a given ``pad_mode``.

    When performing padding, these slices yield the values from the inner
    part of a larger array that are to be assigned to the excess part
    of the same array. Slices for both sides ("left", "right") of
    the arrays in a given ``axis`` are returned.
    """
    # Calculate the start and stop indices for the inner part
    istart_inner = offset[axis]
    n_large = max(lhs_arr.shape[axis], rhs_arr.shape[axis])
    n_small = min(lhs_arr.shape[axis], rhs_arr.shape[axis])
    istop_inner = istart_inner + n_small
    # Number of values padded to left and right
    n_pad_l = istart_inner
    n_pad_r = n_large - istop_inner

    if pad_mode == 'periodic':
        # left: n_pad_l forward, ending at istop_inner - 1
        pad_slc_l = slice(istop_inner - n_pad_l, istop_inner)
        # right: n_pad_r forward, starting at istart_inner
        pad_slc_r = slice(istart_inner, istart_inner + n_pad_r)

    elif pad_mode == 'symmetric':
        # left: n_pad_l backward, ending at istart_inner + 1
        pad_slc_l = slice(istart_inner + n_pad_l, istart_inner, -1)
        # right: n_pad_r backward, starting at istop_inner - 2
        # For the corner case that the stopping index is -1, we need to
        # replace it with None, since -1 as stopping index is equivalent
        # to the last index, which is not what we want (0 as last index).
        istop_r = istop_inner - 2 - n_pad_r
        if istop_r == -1:
            istop_r = None
        pad_slc_r = slice(istop_inner - 2, istop_r, -1)

    elif pad_mode in ('order0', 'order1'):
        # left: only the first entry, using a slice to avoid squeezing
        pad_slc_l = slice(istart_inner, istart_inner + 1)
        # right: only last entry
        pad_slc_r = slice(istop_inner - 1, istop_inner)

    else:
        # Slices are not used, returning trivial ones. The function should not
        # be used for other modes anyway.
        pad_slc_l, pad_slc_r = slice(0), slice(0)

    return pad_slc_l, pad_slc_r


def _apply_padding(lhs_arr, rhs_arr, offset, pad_mode, direction):
    """Apply padding to ``lhs_arr`` according to ``pad_mode``.

    This helper assigns the values in the excess parts (if existent)
    of ``lhs_arr`` according to the provided padding mode.

    This applies to the following values for ``pad_mode``:
    ``periodic``, ``symmetric``, ``order0``, ``order1``

    See `the online documentation
    <https://odlgroup.github.io/odl/math/resizing_ops.html>`_
    on resizing operators for details.
    """
    if pad_mode not in ('periodic', 'symmetric', 'order0', 'order1'):
        return

    full_slc = [slice(None)] * lhs_arr.ndim
    intersec_slc, _ = _intersection_slice_tuples(lhs_arr, rhs_arr, offset)

    if direction == 'forward':
        working_slc = list(intersec_slc)
    else:
        working_slc = list(full_slc)

    # TODO: order axes according to padding size for optimization (largest
    # last)? Axis strides could be important, too.
    for axis, (n_lhs, n_rhs) in enumerate(zip(lhs_arr.shape, rhs_arr.shape)):

        if n_lhs <= n_rhs:
            continue  # restriction, nothing to do

        n_pad_l = offset[axis]
        n_pad_r = n_lhs - n_rhs - n_pad_l

        # Error scenarios with illegal lengths
        if pad_mode == 'order0' and n_rhs == 0:
            raise ValueError('in axis {}: the smaller array must have size '
                             '>= 1 for order 0 padding, got 0'
                             ''.format(axis))

        if pad_mode == 'order1' and n_rhs < 2:
            raise ValueError('in axis {}: the smaller array must have size '
                             '>= 2 for order 1 padding, got {}'
                             ''.format(axis, n_rhs))

        for lr, pad_len in [('left', n_pad_l), ('right', n_pad_r)]:
            if pad_mode == 'periodic' and pad_len > n_rhs:
                raise ValueError('in axis {}: {} padding length {} exceeds '
                                 'the size {} of the smaller array; this is '
                                 'not allowed for periodic padding'
                                 ''.format(axis, lr, pad_len, n_rhs))

            elif pad_mode == 'symmetric' and pad_len >= n_rhs:
                raise ValueError('in axis {}: {} padding length {} is larger '
                                 'or equal to the size {} of the smaller '
                                 'array; this is not allowed for symmetric '
                                 'padding'
                                 ''.format(axis, lr, pad_len, n_rhs))

        # Slice tuples used to index LHS and RHS for left and right padding,
        # respectively. Since `lhs_arr` is used on both sides of the
        # assignments, full slices are used in all axes except `axis`.
        # TODO: change comment

        # TODO: use working_slc instead of full_slc
        lhs_slc_l, lhs_slc_r = list(working_slc), list(working_slc)
        rhs_slc_l, rhs_slc_r = list(working_slc), list(working_slc)

        # We're always using the outer (excess) parts involved in padding
        # on the LHS of the assignment, so we set them here.
        # TODO: change comment
        pad_slc_outer_l, pad_slc_outer_r = _padding_slices_outer(
            lhs_arr, rhs_arr, axis, offset)

        if direction == 'forward':
            lhs_slc_l[axis] = pad_slc_outer_l
            lhs_slc_r[axis] = pad_slc_outer_r
        else:
            rhs_slc_l[axis] = pad_slc_outer_l
            rhs_slc_r[axis] = pad_slc_outer_r

        if pad_mode in ('periodic', 'symmetric'):
            pad_slc_inner_l, pad_slc_inner_r = _padding_slices_inner(
                lhs_arr, rhs_arr, axis, offset, pad_mode)

            # Using `lhs_arr` on both sides of the assignment such that the
            # shapes match and the "corner" blocks are properly assigned
            # or used in the addition for the adjoint, respectively.
            if direction == 'forward':
                rhs_slc_l[axis] = pad_slc_inner_l
                rhs_slc_r[axis] = pad_slc_inner_r

                lhs_arr[tuple(lhs_slc_l)] = lhs_arr[tuple(rhs_slc_l)]
                lhs_arr[tuple(lhs_slc_r)] = lhs_arr[tuple(rhs_slc_r)]
            else:
                lhs_slc_l[axis] = pad_slc_inner_l
                lhs_slc_r[axis] = pad_slc_inner_r

                lhs_arr[tuple(lhs_slc_l)] += lhs_arr[tuple(rhs_slc_l)]
                lhs_arr[tuple(lhs_slc_r)] += lhs_arr[tuple(rhs_slc_r)]

        elif pad_mode == 'order0':
            # The `_padding_slices_inner` helper returns the slices for the
            # boundary values.
            left_slc, right_slc = _padding_slices_inner(
                lhs_arr, rhs_arr, axis, offset, pad_mode)

            if direction == 'forward':
                rhs_slc_l[axis] = left_slc
                rhs_slc_r[axis] = right_slc

                lhs_arr[tuple(lhs_slc_l)] = lhs_arr[tuple(rhs_slc_l)]
                lhs_arr[tuple(lhs_slc_r)] = lhs_arr[tuple(rhs_slc_r)]
            else:
                lhs_slc_l[axis] = left_slc
                lhs_slc_r[axis] = right_slc

                lhs_arr[tuple(lhs_slc_l)] += np.sum(
                    lhs_arr[tuple(rhs_slc_l)],
                    axis=axis, keepdims=True, dtype=lhs_arr.dtype)
                lhs_arr[tuple(lhs_slc_r)] += np.sum(
                    lhs_arr[tuple(rhs_slc_r)],
                    axis=axis, keepdims=True, dtype=lhs_arr.dtype)

        elif pad_mode == 'order1':
            # Some extra work necessary: need to compute the derivative at
            # the boundary and use that to continue with constant derivative.

            # Slice for broadcasting of a 1D array along `axis`
            bcast_slc = [None] * lhs_arr.ndim
            bcast_slc[axis] = slice(None)

            # Slices for the boundary in `axis`
            left_slc, right_slc = _padding_slices_inner(
                lhs_arr, rhs_arr, axis, offset, pad_mode)

            # Create slice tuples for indexing of the boundary values
            bdry_slc_l = list(working_slc)
            bdry_slc_l[axis] = left_slc
            bdry_slc_r = list(working_slc)
            bdry_slc_r[axis] = right_slc

            # For the slope at the boundary, we need two neighboring points.
            # We create the corresponding slices from the boundary slices.
            slope_slc_l = list(working_slc)
            slope_slc_l[axis] = slice(left_slc.start, left_slc.stop + 1)
            slope_slc_r = list(working_slc)
            slope_slc_r[axis] = slice(right_slc.start - 1, right_slc.stop)

            # The `np.arange`s, broadcast along `axis`, are used to create the
            # constant-slope continuation (forward) or to calculate the
            # first order moments (adjoint).
            arange_l = np.arange(-n_pad_l, 0,
                                 dtype=lhs_arr.dtype)[bcast_slc]
            arange_r = np.arange(1, n_pad_r + 1,
                                 dtype=lhs_arr.dtype)[bcast_slc]

            if direction == 'forward':
                # Take first order difference to get the derivative
                # along `axis`.
                slope_l = np.diff(lhs_arr[slope_slc_l], n=1, axis=axis)
                slope_r = np.diff(lhs_arr[slope_slc_r], n=1, axis=axis)

                # Finally assign the constant slope values
                lhs_arr[tuple(lhs_slc_l)] = (
                    lhs_arr[tuple(bdry_slc_l)] + arange_l * slope_l)
                lhs_arr[tuple(lhs_slc_r)] = (
                    lhs_arr[tuple(bdry_slc_r)] + arange_r * slope_r)
            else:
                # Same as in 'order0'
                lhs_arr[tuple(bdry_slc_l)] += np.sum(
                    lhs_arr[tuple(rhs_slc_l)],
                    axis=axis, keepdims=True, dtype=lhs_arr.dtype)
                lhs_arr[tuple(bdry_slc_r)] += np.sum(
                    lhs_arr[tuple(rhs_slc_r)],
                    axis=axis, keepdims=True, dtype=lhs_arr.dtype)

                # Calculate the order 1 moments
                moment1_l = np.sum(arange_l * lhs_arr[tuple(rhs_slc_l)],
                                   axis=axis, keepdims=True,
                                   dtype=lhs_arr.dtype)
                moment1_r = np.sum(arange_r * lhs_arr[tuple(rhs_slc_r)],
                                   axis=axis, keepdims=True,
                                   dtype=lhs_arr.dtype)

                # Add moment1 at the "width-2 boundary layers", with the sign
                # corresponding to the sign in the derivative calculation
                # of the forward padding.
                sign = np.array([-1, 1])[bcast_slc]
                lhs_arr[slope_slc_l] += moment1_l * sign
                lhs_arr[slope_slc_r] += moment1_r * sign

        if direction == 'forward':
            working_slc[axis] = full_slc[axis]
        else:
            working_slc[axis] = intersec_slc[axis]


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Functions for graphical output."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

import numpy as np

from odl.util.testutils import run_doctests
from odl.util.utility import is_real_dtype


__all__ = ('show_discrete_data',)


def _safe_minmax(values):
    """Calculate min and max of array with guards for nan and inf."""

    # Nan and inf guarded min and max
    minval = np.min(values[np.isfinite(values)])
    maxval = np.max(values[np.isfinite(values)])

    return minval, maxval


def _colorbar_ticks(minval, maxval):
    """Return the ticks (values show) in the colorbar."""
    return [minval, (maxval + minval) / 2., maxval]


def _digits(minval, maxval):
    """Digits needed to comforatbly display values in [minval, maxval]"""
    if minval == maxval:
        return 3
    else:
        return min(10, max(2, int(1 + abs(np.log10(maxval - minval)))))


def _colorbar_format(minval, maxval):
    """Return the format string for the colorbar."""
    return '%.{}f'.format(_digits(minval, maxval))


def _axes_info(grid, npoints=5):
    result = []

    min_pt = grid.min()
    max_pt = grid.max()
    for axis in range(grid.ndim):
        xmin = min_pt[axis]
        xmax = max_pt[axis]

        points = np.linspace(xmin, xmax, npoints)
        indices = np.linspace(0, grid.shape[axis] - 1, npoints, dtype=int)
        tick_values = grid.coord_vectors[axis][indices]

        # Do not use corner point in case of a partition, use outer corner
        tick_values[[0, -1]] = xmin, xmax

        format_str = '{:.' + str(_digits(xmin, xmax)) + 'f}'
        tick_labels = [format_str.format(f) for f in tick_values]

        result += [(points, tick_labels)]

    return result


def show_discrete_data(values, grid, title=None, method='',
                       show=False, fig=None, **kwargs):
    """Display a discrete 1d or 2d function.

    Parameters
    ----------
    values : `numpy.ndarray`
        The values to visualize
    grid : `TensorGrid` or `RectPartition`
        Grid of the values

    title : string, optional
        Set the title of the figure

    method : string, optional
        1d methods:

        'plot' : graph plot

        'scatter' : scattered 2d points
        (2nd axis <-> value)

        2d methods:

        'imshow' : image plot with coloring according to value,
        including a colorbar.

        'scatter' : cloud of scattered 3d points
        (3rd axis <-> value)

        'wireframe', 'plot_wireframe' : surface plot


    show : bool, optional
        If the plot should be showed now or deferred until later

    fig : `matplotlib.figure.Figure`
        The figure to show in. Expected to be of same "style", as the figure
        given by this function. The most common usecase is that fig is the
        return value from an earlier call to this function.

    interp : {'nearest', 'linear'}
        Interpolation method to use.

    axis_labels : string
        Axis labels, default: ['x', 'y']

    kwargs : {'figsize', 'saveto', ...}
        Extra keyword arguments passed on to display method
        See the Matplotlib functions for documentation of extra
        options.

    Returns
    -------
    fig : `matplotlib.figure.Figure`
        The resulting figure. It is also shown to the user.
    colorbar : `matplotlib.colorbar.Colorbar`
        The colorbar

    See Also
    --------
    matplotlib.pyplot.plot : Show graph plot

    matplotlib.pyplot.imshow : Show data as image

    matplotlib.pyplot.scatter : Show scattered 3d points
    """
    # Importing pyplot takes ~2 sec, only import when needed.
    import matplotlib.pyplot as plt

    args_re = []
    args_im = []
    dsp_kwargs = {}
    sub_kwargs = {}
    arrange_subplots = (121, 122)  # horzontal arrangement

    # Create axis labels which remember their original meaning
    axis_labels = kwargs.pop('axis_labels', ['x', 'y'])

    values_are_complex = not is_real_dtype(values.dtype)
    figsize = kwargs.pop('figsize', None)
    saveto = kwargs.pop('saveto', None)
    interp = kwargs.pop('interp', 'nearest')

    if values.ndim == 1:  # TODO: maybe a plotter class would be better
        if not method:
            if interp == 'nearest':
                method = 'step'
                dsp_kwargs['where'] = 'mid'
            elif interp == 'linear':
                method = 'plot'
            else:
                method = 'plot'

        if method == 'plot' or method == 'step' or method == 'scatter':
            args_re += [grid.coord_vectors[0], values.real]
            args_im += [grid.coord_vectors[0], values.imag]
        else:
            raise ValueError('`method` {!r} not supported'
                             ''.format(method))

    elif values.ndim == 2:
        if not method:
            method = 'imshow'

        if method == 'imshow':
            args_re = [np.rot90(values.real)]
            args_im = [np.rot90(values.imag)] if values_are_complex else []

            extent = [grid.min()[0], grid.max()[0],
                      grid.min()[1], grid.max()[1]]

            if interp == 'nearest':
                interpolation = 'nearest'
            elif interp == 'linear':
                interpolation = 'bilinear'
            else:
                interpolation = 'none'

            dsp_kwargs.update({'interpolation': interpolation,
                               'cmap': 'bone',
                               'extent': extent,
                               'aspect': 'auto'})
        elif method == 'scatter':
            pts = grid.points()
            args_re = [pts[:, 0], pts[:, 1], values.ravel().real]
            args_im = ([pts[:, 0], pts[:, 1], values.ravel().imag]
                       if values_are_complex else [])
            sub_kwargs.update({'projection': '3d'})
        elif method in ('wireframe', 'plot_wireframe'):
            method = 'plot_wireframe'
            x, y = grid.meshgrid
            args_re = [x, y, np.rot90(values.real)]
            args_im = ([x, y, np.rot90(values.imag)] if values_are_complex
                       else [])
            sub_kwargs.update({'projection': '3d'})
        else:
            raise ValueError('`method` {!r} not supported'
                             ''.format(method))

    else:
        raise NotImplementedError('no method for {}d display implemented'
                                  ''.format(values.ndim))

    # Additional keyword args are passed on to the display method
    dsp_kwargs.update(**kwargs)

    if fig is not None:
        # Reuse figure if given as input
        if not isinstance(fig, plt.Figure):
            raise TypeError('`fig` {} not a matplotlib figure'.format(fig))

        if not plt.fignum_exists(fig.number):
            # If figure does not exist, user either closed the figure or
            # is using IPython, in this case we need a new figure.

            fig = plt.figure(figsize=figsize)
            updatefig = False
        else:
            # Set current figure to given input
            fig = plt.figure(fig.number)
            updatefig = True

            if values.ndim > 1:
                # If the figure is larger than 1d, we can clear it since we
                # dont reuse anything. Keeping it causes performance problems.
                fig.clf()
    else:
        fig = plt.figure(figsize=figsize)
        updatefig = False

    if values_are_complex:
        # Real
        if len(fig.axes) == 0:
            # Create new axis if needed
            sub_re = plt.subplot(arrange_subplots[0], **sub_kwargs)
            sub_re.set_title('Real part')
            sub_re.set_xlabel(axis_labels[0])
            if values.ndim == 2:
                sub_re.set_ylabel(axis_labels[1])
            else:
                sub_re.set_ylabel('value')
        else:
            sub_re = fig.axes[0]

        display_re = getattr(sub_re, method)
        csub_re = display_re(*args_re, **dsp_kwargs)

        # Axis ticks
        if method == 'imshow' and not grid.is_uniform:
            (xpts, xlabels), (ypts, ylabels) = _axes_info(grid)
            plt.xticks(xpts, xlabels)
            plt.yticks(ypts, ylabels)

        if method == 'imshow' and len(fig.axes) < 2:
            # Create colorbar if none seems to exist

            # Use clim from kwargs if given
            if 'clim' not in kwargs:
                minval_re, maxval_re = _safe_minmax(values.real)
            else:
                minval_re, maxval_re = kwargs['clim']

            ticks_re = _colorbar_ticks(minval_re, maxval_re)
            format_re = _colorbar_format(minval_re, maxval_re)

            plt.colorbar(csub_re, orientation='horizontal',
                         ticks=ticks_re, format=format_re)

        # Imaginary
        if len(fig.axes) < 3:
            sub_im = plt.subplot(arrange_subplots[1], **sub_kwargs)
            sub_im.set_title('Imaginary part')
            sub_im.set_xlabel(axis_labels[0])
            if values.ndim == 2:
                sub_im.set_ylabel(axis_labels[1])
            else:
                sub_im.set_ylabel('value')
        else:
            sub_im = fig.axes[2]

        display_im = getattr(sub_im, method)
        csub_im = display_im(*args_im, **dsp_kwargs)

        # Axis ticks
        if method == 'imshow' and not grid.is_uniform:
            (xpts, xlabels), (ypts, ylabels) = _axes_info(grid)
            plt.xticks(xpts, xlabels)
            plt.yticks(ypts, ylabels)

        if method == 'imshow' and len(fig.axes) < 4:
            # Create colorbar if none seems to exist

            # Use clim from kwargs if given
            if 'clim' not in kwargs:
                minval_im, maxval_im = _safe_minmax(values.imag)
            else:
                minval_im, maxval_im = kwargs['clim']

            ticks_im = _colorbar_ticks(minval_im, maxval_im)
            format_im = _colorbar_format(minval_im, maxval_im)

            plt.colorbar(csub_im, orientation='horizontal',
                         ticks=ticks_im, format=format_im)

    else:
        if len(fig.axes) == 0:
            # Create new axis object if needed
            sub = plt.subplot(111, **sub_kwargs)
            sub.set_xlabel(axis_labels[0])
            if values.ndim == 2:
                sub.set_ylabel(axis_labels[1])
            else:
                sub.set_ylabel('value')
            try:
                # For 3d plots
                sub.set_zlabel('z')
            except AttributeError:
                pass
        else:
            sub = fig.axes[0]

        display = getattr(sub, method)
        csub = display(*args_re, **dsp_kwargs)

        # Axis ticks
        if method == 'imshow' and not grid.is_uniform:
            (xpts, xlabels), (ypts, ylabels) = _axes_info(grid)
            plt.xticks(xpts, xlabels)
            plt.yticks(ypts, ylabels)

        if method == 'imshow' and len(fig.axes) < 2:
            # Create colorbar if none seems to exist

            # Use clim from kwargs if given
            if 'clim' not in kwargs:
                minval, maxval = _safe_minmax(values)
            else:
                minval, maxval = kwargs['clim']

            ticks = _colorbar_ticks(minval, maxval)
            format = _colorbar_format(minval, maxval)

            plt.colorbar(mappable=csub, ticks=ticks, format=format)

    # Fixes overlapping stuff at the expense of potentially squashed subplots
    fig.tight_layout()

    if title is not None:
        if not values_are_complex:
            # Do not overwrite title for complex values
            plt.title(title)
        fig.canvas.manager.set_window_title(title)

    if updatefig or plt.isinteractive():
        # If we are running in interactive mode, we can always show the fig
        # This causes an artifact, where users of `CallbackShow` without
        # interactive mode only shows the figure after the second iteration.
        plt.show(block=False)
        plt.draw()
        plt.pause(0.1)

    if show:
        plt.show()

    if saveto is not None:
        fig.savefig(saveto)
    return fig


if __name__ == '__main__':
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Utilities for creating phantoms."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()

import numpy as np


__all__ = ('cylinders_from_ellipses',)


def cylinders_from_ellipses(ellipses2d):
    """Create 3d cylinders from ellipses."""
    ellipses2d = np.asarray(ellipses2d)
    ellipses3d = np.zeros((ellipses2d.shape[0], 10))
    ellipses3d[:, [0, 1, 2, 4, 5, 7]] = ellipses2d
    ellipses3d[:, 3] = 100000.0

    return ellipses3d


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Fan beam geometries in 2 dimensions."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

import numpy as np

from odl.tomo.geometry.detector import Flat1dDetector
from odl.tomo.geometry.geometry import DivergentBeamGeometry
from odl.tomo.util.utility import euler_matrix, perpendicular_vector


__all__ = ('FanFlatGeometry',)


class FanFlatGeometry(DivergentBeamGeometry):

    """Abstract 2d fan beam geometry with flat 1d detector.

    The source moves on a circle with radius ``src_radius``, and the
    detector reference point is opposite to the source, i.e. at maximum
    distance, on a circle with radius ``det_radius``. One of the two
    radii can be chosen as 0, which corresponds to a stationary source
    or detector, respectively.

    The motion parameter is the 1d rotation angle parameterizing source
    and detector positions simultaneously.

    In the standard configuration, the source and detector start on the
    first coodinate axis with vector ``(1, 0)`` from source to detector,
    and the initial detector axis is ``(0, 1)``.
    """

    def __init__(self, apart, dpart, src_radius, det_radius, **kwargs):
        """Initialize a new instance.

        Parameters
        ----------
        apart : 1-dim. `RectPartition`
            Partition of the angle interval
        dpart : 1-dim. `RectPartition`
            Partition of the detector parameter interval
        src_radius : nonnegative float
            Radius of the source circle
        det_radius : nonnegative float
            Radius of the detector circle
        src_to_det_init : `array-like` (shape ``(2,)``), optional
            Initial state of the vector pointing from source to detector
            reference point. The zero vector is not allowed.
            Default: ``(1, 0)``.
        det_init_axis : `array-like` (shape ``(2,)``), optional
            Initial axis defining the detector orientation.
            By default, a normalized `perpendicular_vector` to
            ``src_to_det_init`` is used.
        """
        src_to_det_init = kwargs.pop('src_to_det_init', (1.0, 0.0))
        det_init_axis = kwargs.pop('det_init_axis', None)

        if np.shape(src_to_det_init) != (2,):
            raise ValueError('`src_to_det_init` has shape {}, '
                             'expected (2,)'
                             ''.format(np.shape(src_to_det_init)))
        if np.linalg.norm(src_to_det_init) <= 1e-10:
            raise ValueError('`src_to_det_init` {} too close '
                             'to zero'.format(src_to_det_init))
        self._src_to_det_init = (np.asarray(src_to_det_init, dtype='float64') /
                                 np.linalg.norm(src_to_det_init))

        if det_init_axis is None:
            det_init_axis = perpendicular_vector(self._src_to_det_init)

        self._src_radius, src_radius_in = float(src_radius), src_radius
        if self.src_radius < 0:
            raise ValueError('`src_radius` {} is negative'
                             ''.format(src_radius_in))
        self._det_radius, det_radius_in = float(det_radius), det_radius
        if det_radius < 0:
            raise ValueError('`det_radius` {} is negative'
                             ''.format(det_radius_in))
        if self.src_radius == 0 and self.det_radius == 0:
            raise ValueError('source and detector radii cannot both be 0')

        detector = Flat1dDetector(dpart, det_init_axis)
        super().__init__(ndim=2, motion_part=apart, detector=detector)

    @property
    def src_radius(self):
        """Source circle radius of this geometry."""
        return self._src_radius

    @property
    def det_radius(self):
        """Detector circle radius of this geometry."""
        return self._det_radius

    @property
    def angles(self):
        """Discrete angles given in this geometry."""
        return self.motion_grid.coord_vectors[0]

    def src_position(self, angle):
        """Return the source position at ``angle``.

        For an angle ``phi``, the source position is given by::

            src(phi) = -src_rad * rot_matrix(phi) * src_to_det_init

        where ``src_to_det_init`` is the initial unit vector pointing
        from source to detector.

        Parameters
        ----------
        angle : float
            Rotation angle given in radians, must be contained in
            this geometry's `motion_params`

        Returns
        -------
        point : `numpy.ndarray`, shape ``(2,)``
            Source position corresponding to the given angle
        """
        if angle not in self.motion_params:
            raise ValueError('`angle` {} is not in the valid range {}'
                             ''.format(angle, self.motion_params))

        # Initial vector from 0 to the source. It can be computed this way
        # since source and detector are at maximum distance, i.e. the
        # connecting line passes the origin.
        origin_to_src_init = -self.src_radius * self._src_to_det_init
        return self.rotation_matrix(angle).dot(origin_to_src_init)

    def det_refpoint(self, angle):
        """Return the detector reference point position at ``angle``.

        For an angle ``phi``, the detector position is given by::

            ref(phi) = det_rad * rot_matrix(phi) * src_to_det_init

        where ``src_to_det_init`` is the initial unit vector pointing
        from source to detector.

        Parameters
        ----------
        angle : float
            Rotation angle given in radians, must be contained in
            this geometry's `motion_params`

        Returns
        -------
        point : `numpy.ndarray`, shape (2,)
            Detector reference point corresponding to the given angle

        See Also
        --------
        rotation_matrix
        """
        if angle not in self.motion_params:
            raise ValueError('`angle` {} is not in the valid range {}'
                             ''.format(angle, self.motion_params))

        # Initial vector from 0 to the detector. It can be computed this way
        # since source and detector are at maximum distance, i.e. the
        # connecting line passes the origin.
        origin_to_det_init = self.det_radius * self._src_to_det_init
        return self.rotation_matrix(angle).dot(origin_to_det_init)

    def rotation_matrix(self, angle):
        """Return the rotation matrix for ``angle``.

        For an angle ``phi``, the matrix is given by::

            rot(phi) = [[cos(phi), -sin(phi)],
                        [sin(phi), cos(phi)]]

        Parameters
        ----------
        angle : float
            Rotation angle given in radians, must be contained in
            this geometry's `motion_params`

        Returns
        -------
        rot : `numpy.ndarray`, shape (2, 2)
            The rotation matrix mapping the standard basis vectors in
            the fixed ("lab") coordinate system to the basis vectors of
            the local coordinate system of the detector reference point,
            expressed in the fixed system
        """
        angle = float(angle)
        if angle not in self.motion_params:
            raise ValueError('`angle` {} not in the valid range {}'
                             ''.format(angle, self.motion_params))
        return euler_matrix(angle)

    # TODO: back projection weighting function?

    def __repr__(self):
        """Return ``repr(self)``."""
        arg_fstr = '{!r}, {!r}, src_radius={}, det_radius={}'

        if not np.allclose(self._src_to_det_init, [1, 0]):
            arg_fstr += ',\n    src_to_det_init={src_to_det_init!r}'

        default_axis = perpendicular_vector(self._src_to_det_init)
        if not np.allclose(self.detector.axis, default_axis):
            arg_fstr += ',\n    det_init_axis={det_init_axis!r}'

        arg_str = arg_fstr.format(self.motion_partition, self.det_partition,
                                  self.src_radius, self.det_radius,
                                  src_to_det_init=self._src_to_det_init,
                                  det_init_axis=self.detector.axis)

        return '{}({})'.format(self.__class__.__name__, arg_str)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Detectors for tomographic imaging."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import object, super

from abc import ABCMeta, abstractmethod
import numpy as np

from odl.discr import RectPartition
from odl.tomo.util.utility import perpendicular_vector
from odl.util.utility import with_metaclass


__all__ = ('Detector', 'FlatDetector', 'Flat1dDetector', 'Flat2dDetector',
           'CircleSectionDetector')


class Detector(with_metaclass(ABCMeta, object)):

    """Abstract detector class.

    A detector is described by

    * a set of parameters for surface parametrization (including sampling),
    * a function mapping a surface parameter to the location of a detector
      point relative to its reference point,
    * optionally a surface measure function.
    """

    def __init__(self, part):
        """Initialize a new instance.

        Parameters
        ----------
        part : `RectPartition`
           Partition of the detector parameter set (pixelization).
           It determines dimension, parameter range and discretization.
        """
        if not isinstance(part, RectPartition):
            raise TypeError('`part` {!r} is not a RectPartition instance'
                            ''.format(part))

        self._part = part

    @abstractmethod
    def surface(self, param):
        """Parametrization of the detector reference surface.

        Parameters
        ----------
        param : `params` element
            Parameter value where to evaluate the function

        Returns
        -------
        point :
            Spatial location of the detector point corresponding to
            ``param``
        """

    @property
    def partition(self):
        """Partition of the detector parameter set into subsets."""
        return self._part

    @property
    def ndim(self):
        """Number of dimensions of this detector (0, 1 or 2)."""
        return self.partition.ndim

    @property
    def params(self):
        """Surface parameter set of this detector."""
        return self.partition.set

    @property
    def grid(self):
        """Sampling grid of the parameters."""
        return self.partition.grid

    @property
    def shape(self):
        """Number of subsets (pixels) of the detector per axis."""
        return self.partition.shape

    @property
    def size(self):
        """Total number of pixels."""
        return self.partition.size

    def surface_deriv(self, param):
        """Partial derivative(s) of the surface parametrization.

        Parameters
        ----------
        param : `params` element
            The parameter value where to evaluate the function

        Returns
        -------
        deriv :
            Vector (``ndim=1``) or sequence of vectors corresponding
            to the partial derivatives at ``param``
        """
        raise NotImplementedError

    def surface_measure(self, param):
        """Density function of the surface measure.

        This is the default implementation relying on the `surface_deriv`
        method. For ``ndim == 1``, the density is given by the `Arc
        length`_, for ``ndim == 2``, it is the length of the cross product
        of the partial derivatives of the parametrization, see Wikipedia's
        `Surface area`_ article.

        Parameters
        ----------
        param : `params` element
            The parameter value where to evaluate the function

        Returns
        -------
        measure : float
            The density value at the given parameter

        .. _Arc length:
            https://en.wikipedia.org/wiki/Curve#Lengths_of_curves
        .. _Surface area:
            https://en.wikipedia.org/wiki/Surface_area
        """
        if param not in self.params:
            raise ValueError('`param` {} not in the valid range {}'
                             ''.format(param, self.params))
        if self.ndim == 1:
            return float(np.linalg.norm(self.surface_deriv(param)))
        elif self.ndim == 2:
            return float(np.linalg.norm(np.cross(*self.surface_deriv(param))))
        else:
            raise NotImplementedError


class FlatDetector(Detector):

    """Abstract class for flat detectors in 2 and 3 dimensions."""

    def surface_measure(self, param=None):
        """Constant density function of the surface measure.

        Parameters
        ----------
        param : `params` element, optional
            Parameter value where to evaluate the function

        Returns
        -------
        measure : float
            Constant density 1.0
        """
        if param not in self.params:
            raise ValueError('`param` {} not in the valid range '
                             '{}'.format(param, self.params))
        # TODO: apart from being constant, there is no big simplification
        # in this method compared to parent. Consider removing FlatDetector
        # altogether.
        return super().surface_measure(self.params.min_pt)


class Flat1dDetector(FlatDetector):

    """A 1d line detector aligned with ``axis``."""

    def __init__(self, part, axis):
        """Initialize a new instance.

        Parameters
        ----------
        part : 1-dim. `RectPartition`
            Partition of the parameter interval, corresponding to the
            line elements
        axis : `array-like`, shape ``(2,)``
            Principal axis of the detector
        """
        super().__init__(part)
        if self.ndim != 1:
            raise ValueError('expected partition to have 1 dimension, '
                             'got {}'.format(self.ndim))

        if np.linalg.norm(axis) <= 1e-10:
            raise ValueError('`axis` vector {} too close to zero'
                             ''.format(axis))
        self._axis = np.asarray(axis) / np.linalg.norm(axis)
        self._normal = perpendicular_vector(self.axis)

    @property
    def axis(self):
        """Normalized principal axis of the detector."""
        return self._axis

    @property
    def normal(self):
        """Unit vector perpendicular to the detector.

        Its orientation is chosen such that the system ``axis, normal``
        is right-handed.
        """
        return self._normal

    def surface(self, param):
        """Parametrization of the (1d) detector reference surface.

        The reference line segment is chosen to be aligned with the
        second coordinate axis, such that the parameter value 0 results
        in the reference point (0, 0).

        Parameters
        ----------
        param : `params` element
            The parameter value where to evaluate the function

        Returns
        -------
        point : `numpy.ndarray`, shape (2,)
            The point on the detector surface corresponding to the
            given parameters
        """
        param = float(param)
        if param not in self.params:
            raise ValueError('`param` {} not in the valid range '
                             '{}'.format(param, self.params))
        return self.axis * param

    def surface_deriv(self, param=None):
        """Derivative of the surface parametrization.

        Parameters
        ----------
        param : `params` element, optional
            The parameter value where to evaluate the function

        Returns
        -------
        derivative : `numpy.ndarray`, shape (2,)
            The constant derivative
        """
        if param is not None and param not in self.params:
            raise ValueError('`param` {} not in the valid range '
                             '{}'.format(param, self.params))
        return self.axis

    def __repr__(self):
        """Return ``repr(self)``."""
        inner_fstr = '\n    {!r},\n    {!r}'
        inner_str = inner_fstr.format(self.partition, self.axis)
        return '{}({})'.format(self.__class__.__name__, inner_str)

    def __str__(self):
        """Return ``str(self)``."""
        # TODO: prettify
        inner_fstr = '\n    {},\n    {}'
        inner_str = inner_fstr.format(self.partition, self.axis)
        return '{}({})'.format(self.__class__.__name__, inner_str)


class Flat2dDetector(FlatDetector):

    """A 2d flat panel detector aligned with ``axes``."""

    def __init__(self, part, axes):
        """Initialize a new instance.

        Parameters
        ----------
        part : 1-dim. `RectPartition`
            Partition of the parameter interval, corresponding to the
            pixels
        axes : 2-tuple of `array-like`'s (shape ``(3,)``)
            Principal axes of the detector, e.g.
            ``[(0, 1, 0), (0, 0, 1)]``
        """
        super().__init__(part)
        if self.ndim != 2:
            raise ValueError('expected partition to have 2 dimensions, '
                             'got {}'.format(self.ndim))

        for i, a in enumerate(axes):
            if np.linalg.norm(a) <= 1e-10:
                raise ValueError('axis vector {} {} too close to zero'
                                 ''.format(i, axes[i]))
            if np.shape(a) != (3,):
                raise ValueError('axis vector {} has shape {}. expected (3,)'
                                 ''.format(i, np.shape(a)))

        self._axes = tuple(np.asarray(a) / np.linalg.norm(a) for a in axes)
        self._normal = np.cross(self.axes[0], self.axes[1])

        if np.linalg.norm(self.normal) <= 1e-4:
            raise ValueError('`axes` are almost parallel (norm of normal = '
                             '{})'.format(np.linalg.norm(self.normal)))

    @property
    def axes(self):
        """Normalized principal axes of this detector as a 2-tuple."""
        return self._axes

    @property
    def normal(self):
        """Unit vector perpendicular to this detector.

        The orientation is chosen such that the triple
        ``axes[0], axes[1], normal`` form a right-hand system.
        """
        return self._normal

    def surface(self, param):
        """Parametrization of the 2d detector reference surface.

        The reference plane segment is chosen to be aligned with the
        second and third coordinate axes, in this order, such that
        the parameter value (0, 0) results in the reference (0, 0, 0).

        Parameters
        ----------
        param : `params` element
            The parameter value where to evaluate the function

        Returns
        -------
        point : `numpy.ndarray`, shape (3,)
            The point on the detector surface corresponding to the
            given parameters
        """
        if param not in self.params:
            raise ValueError('`param` {} not in the valid range '
                             '{}'.format(param, self.params))

        return sum(float(p) * ax for p, ax in zip(param, self.axes))

    def surface_deriv(self, param=None):
        """Derivative of the surface parametrization.

        Parameters
        ----------
        param : `params` element, optional
            The parameter value where to evaluate the function

        Returns
        -------
        derivatives : 2-tuple of `numpy.ndarray`'s (shape ``(3,)``)
            The constant partial derivatives given by the detector axes
        """
        if param is not None and param not in self.params:
            raise ValueError('`param` {} not in the valid range '
                             '{}'.format(param, self.params))
        return self.axes

    def __repr__(self):
        """Return ``repr(self)``."""
        inner_fstr = '\n    {!r},\n    {!r}'
        inner_str = inner_fstr.format(self.partition, self.axes)
        return '{}({})'.format(self.__class__.__name__, inner_str)

    def __str__(self):
        """Return ``str(self)``."""
        # TODO: prettify
        inner_fstr = '\n    {},\n    {}'
        inner_str = inner_fstr.format(self.partition, self.axes)
        return '{}({})'.format(self.__class__.__name__, inner_str)


class CircleSectionDetector(Detector):

    """A 1d detector given by a section of a circle.

    The reference circular section is part of a circle with radius ``r``,
    which is shifted by the vector ``(-r, 0)`` such that the parameter
    value 0 results in the detector reference point ``(0, 0)``.
    """

    def __init__(self, part, circ_rad):
        """Initialize a new instance.

        Parameters
        ----------
        part : 1-dim. `RectPartition`
            Partition of the parameter interval, corresponding to the
            angle sections along the line
        circ_rad : positive float
            Radius of the circle along which the detector is curved
        """
        super().__init__(part)
        if self.ndim != 1:
            raise ValueError('expected `part` to have 1 dimension, '
                             'got {}'.format(self.ndim))

        self._circ_rad, circ_rad_in = float(circ_rad), circ_rad
        if self.circ_rad <= 0:
            raise ValueError('`circ_rad` {} is not positive'
                             ''.format(circ_rad_in))

    @property
    def circ_rad(self):
        """Circle radius of this detector."""
        return self._circ_rad

    def surface(self, param):
        """Parametrization of the detector reference surface.

        Parameters
        ----------
        param : `params` element
            The parameter value where to evaluate the function
        """
        if param in self.params or self.params.contains_all(param):
            return (self.circ_rad *
                    np.array([np.cos(param) - 1, np.sin(param)]).T)
        else:
            raise ValueError('`param` value(s) {} not in the valid range '
                             '{}'.format(param, self.params))

    def surface_deriv(self, param):
        """Partial derivative(s) of the surface parametrization.

        Parameters
        ----------
        param : `params` element
            The parameter value where to evaluate the function
        """
        if param in self.params or self.params.contains_all(param):
            return self.circ_rad * np.array([-np.sin(param), np.cos(param)]).T
        else:
            raise ValueError('`param` value(s) {} not in the valid range '
                             '{}'.format(param, self.params))

    def surface_measure(self, param):
        """Constant density function of the surface measure.

        Parameters
        ----------
        param : `params` element
            The parameter value where to evaluate the function

        Returns
        -------
        measure : float
            The constant density ``r``, equal to the length of the
            tangent to the detector circle at any point
        """
        if param in self.params:
            return self.circ_rad
        elif self.params.contains_all(param):
            return self.circ_rad * np.ones_like(param, dtype=float)
        else:
            raise ValueError('`param` value(s) {} not in the valid range '
                             '{}'.format(param, self.params))

    def __repr__(self):
        """Return ``repr(self)``."""
        inner_fstr = '\n    {!r},\n    {}'
        inner_str = inner_fstr.format(self.partition, self.circ_rad)
        return '{}({})'.format(self.__class__.__name__, inner_str)

    def __str__(self):
        """Return ``str(self)``."""
        # TODO: prettify
        inner_fstr = '\n    {},\n    {}'
        inner_str = inner_fstr.format(self.partition, self.circ_rad)
        return '{}({})'.format(self.__class__.__name__, inner_str)


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Operators defined for tensor fields."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

import numpy as np

from odl.discr.lp_discr import DiscreteLp
from odl.discr.tensor_ops import PointwiseTensorFieldOperator
from odl.space import ProductSpace


__all__ = ('PartialDerivative', 'Gradient', 'Divergence', 'Laplacian')

_SUPPORTED_DIFF_METHODS = ('central', 'forward', 'backward')
_SUPPORTED_PAD_MODES = ('constant',
                        'symmetric', 'symmetric_adjoint',
                        'periodic',
                        'order0', 'order0_adjoint',
                        'order1', 'order1_adjoint',
                        'order2', 'order2_adjoint')

_ADJ_METHOD = {'central': 'central',
               'forward': 'backward',
               'backward': 'forward'}

_ADJ_PADDING = {'constant': 'constant',
                'symmetric': 'symmetric_adjoint',
                'symmetric_adjoint': 'symmetric',
                'periodic': 'periodic',
                'order0': 'order0_adjoint',
                'order0_adjoint': 'order0',
                'order1': 'order1_adjoint',
                'order1_adjoint': 'order1',
                'order2': 'order2_adjoint',
                'order2_adjoint': 'order2'}


class PartialDerivative(PointwiseTensorFieldOperator):

    """Calculate the discrete partial derivative along a given axis.

    Calls helper function `finite_diff` to calculate finite difference.
    Preserves the shape of the underlying grid.
    """

    def __init__(self, space, axis, method='forward', pad_mode='constant',
                 pad_const=0):
        """Initialize a new instance.

        Parameters
        ----------
        space : `DiscreteLp`
            Space of elements which the operator is acting on.
        axis : int, optional
            Axis along which the partial derivative is evaluated.
        method : {'central', 'forward', 'backward'}, optional
            Finite difference method which is used in the interior of the
            domain of ``f``.
        pad_mode : string, optional
            The padding mode to use outside the domain.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'order2'``: Extend with second order accuracy (ensures continuity
            of the second derivative). This requires at least 3 values along
            the ``axis`` where padding is applied.

        pad_const : float, optional
            For ``pad_mode == 'constant'``, ``f`` assumes
            ``pad_const`` for indices outside the domain of ``f``

        Examples
        --------
        >>> f = np.array([[ 0.,  1.,  2.,  3.,  4.],
        ...               [ 0.,  2.,  4.,  6.,  8.]])
        >>> discr = odl.uniform_discr([0, 0], [2, 1], f.shape)
        >>> par_deriv = PartialDerivative(discr, axis=0, pad_mode='order1')
        >>> par_div_f = par_deriv(f)
        >>> print(par_div_f)
        [[0.0, 1.0, 2.0, 3.0, 4.0],
         [0.0, 1.0, 2.0, 3.0, 4.0]]
        """
        if not isinstance(space, DiscreteLp):
            raise TypeError('`space` {!r} is not a DiscreteLp instance'
                            ''.format(space))

        # Method is affine if nonzero padding is given.
        linear = not (pad_mode == 'constant' and pad_const != 0)
        super().__init__(domain=space, range=space, linear=linear)
        self.axis = int(axis)
        self.dx = space.cell_sides[axis]

        self.method, method_in = str(method).lower(), method
        if method not in _SUPPORTED_DIFF_METHODS:
            raise ValueError('`method` {} not understood'
                             ''.format(method_in))

        self.pad_mode, pad_mode_in = str(pad_mode).lower(), pad_mode
        if pad_mode not in _SUPPORTED_PAD_MODES:
            raise ValueError('`pad_mode` {} not understood'
                             ''.format(pad_mode_in))

        self.pad_const = space.field.element(pad_const)

    def _call(self, x, out=None):
        """Calculate partial derivative of ``x``."""
        if out is None:
            out = self.range.element()

        # TODO: this pipes CUDA arrays through NumPy. Write native operator.
        out_arr = out.asarray()
        finite_diff(x.asarray(), out=out_arr, axis=self.axis, dx=self.dx,
                    method=self.method, pad_mode=self.pad_mode,
                    pad_const=self.pad_const)

        # self assignment: no overhead in the case out_arr is a view
        out[:] = out_arr
        return out

    def derivative(self, point=None):
        """Return the derivative operator.

        The partial derivative is usually linear, but in case the 'constant'
        ``pad_mode`` is used with nonzero ``pad_const``, the
        derivative is given by the derivative with 0 ``pad_const``.

        Parameters
        ----------
        point : `domain` `element-like`, optional
            The point to take the derivative in. Does not change the result
            since the operator is affine.
        """
        if self.pad_mode == 'constant' and self.pad_const != 0:
            return PartialDerivative(self.domain, self.axis, self.method,
                                     self.pad_mode, 0)
        else:
            return self

    @property
    def adjoint(self):
        """Return the adjoint operator."""
        if not self.is_linear:
            raise ValueError('operator with nonzero pad_const ({}) is not'
                             ' linear and has no adjoint'
                             ''.format(self.pad_const))

        return -PartialDerivative(self.domain, self.axis,
                                  _ADJ_METHOD[self.method],
                                  _ADJ_PADDING[self.pad_mode],
                                  self.pad_const)


class Gradient(PointwiseTensorFieldOperator):

    """Spatial gradient operator for `DiscreteLp` spaces.

    Calls helper function `finite_diff` to calculate each component of the
    resulting product space element. For the adjoint of the `Gradient`
    operator, zero padding is assumed to match the negative `Divergence`
    operator
    """

    def __init__(self, domain=None, range=None, method='forward',
                 pad_mode='constant', pad_const=0):
        """Initialize a new instance.

        Zero padding is assumed for the adjoint of the `Gradient`
        operator to match negative `Divergence` operator.

        Parameters
        ----------
        domain : `DiscreteLp`, optional
            Space of elements which the operator acts on.
            This is required if ``range`` is not given.
        range : power space of `DiscreteLp`, optional
            Space of elements to which the operator maps.
            This is required if ``domain`` is not given.
        method : {'central', 'forward', 'backward'}, optional
            Finite difference method to be used
        pad_mode : string, optional
            The padding mode to use outside the domain.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis where padding is applied.

            ``'order2'``: Extend with second order accuracy (ensures continuity
            of the second derivative). This requires at least 3 values along
            each axis.

        pad_const : float, optional
            For ``pad_mode == 'constant'``, ``f`` assumes
            ``pad_const`` for indices outside the domain of ``f``

        Examples
        --------
        Creating a Gradient operator:

        >>> dom = odl.uniform_discr([0, 0], [1, 1], (10, 20))
        >>> ran = odl.ProductSpace(dom, dom.ndim)  # 2-dimensional
        >>> grad_op = Gradient(dom)
        >>> grad_op.range == ran
        True
        >>> grad_op2 = Gradient(range=ran)
        >>> grad_op2.domain == dom
        True
        >>> grad_op3 = Gradient(domain=dom, range=ran)
        >>> grad_op3.domain == dom
        True
        >>> grad_op3.range == ran
        True

        Calling the operator:

        >>> data = np.array([[ 0., 1., 2., 3., 4.],
        ...                  [ 0., 2., 4., 6., 8.]])
        >>> discr = odl.uniform_discr([0, 0], [2, 5], data.shape)
        >>> f = discr.element(data)
        >>> grad = Gradient(discr)
        >>> grad_f = grad(f)
        >>> print(grad_f[0])
        [[0.0, 1.0, 2.0, 3.0, 4.0],
         [0.0, -2.0, -4.0, -6.0, -8.0]]
        >>> print(grad_f[1])
        [[1.0, 1.0, 1.0, 1.0, -4.0],
         [2.0, 2.0, 2.0, 2.0, -8.0]]

        Verify adjoint:

        >>> g = grad.range.element((data, data ** 2))
        >>> adj_g = grad.adjoint(g)
        >>> print(adj_g)
        [[0.0, -2.0, -5.0, -8.0, -11.0],
         [0.0, -5.0, -14.0, -23.0, -32.0]]
        >>> g.inner(grad_f) / f.inner(adj_g)
        1.0
        """
        if domain is None and range is None:
            raise ValueError('either `domain` or `range` must be specified')

        if domain is None:
            if not isinstance(range, ProductSpace):
                raise TypeError('`range` {!r} is not a ProductSpace instance'
                                ''.format(range))
            domain = range[0]

        if range is None:
            if not isinstance(domain, DiscreteLp):
                raise TypeError('`domain` {!r} is not a `DiscreteLp` '
                                'instance'.format(domain))
            range = ProductSpace(domain, domain.ndim)

        linear = not (pad_mode == 'constant' and pad_const != 0)
        super().__init__(domain, range, linear=linear)

        self.method, method_in = str(method).lower(), method
        if method not in _SUPPORTED_DIFF_METHODS:
            raise ValueError('`method` {} not understood'
                             ''.format(method_in))

        self.pad_mode, pad_mode_in = str(pad_mode).lower(), pad_mode
        if pad_mode not in _SUPPORTED_PAD_MODES:
            raise ValueError('`pad_mode` {} not understood'
                             ''.format(pad_mode_in))

        self.pad_const = domain.field.element(pad_const)

    def _call(self, x, out=None):
        """Calculate the spatial gradient of ``x``."""
        if out is None:
            out = self.range.element()

        x_arr = x.asarray()
        ndim = self.domain.ndim
        dx = self.domain.cell_sides

        for axis in range(ndim):
            out_arr = out[axis].asarray()

            finite_diff(x_arr, axis=axis, dx=dx[axis], method=self.method,
                        pad_mode=self.pad_mode,
                        pad_const=self.pad_const,
                        out=out_arr)

            out[axis][:] = out_arr

        return out

    def derivative(self, point=None):
        """Return the derivative operator.

        The gradient is usually linear, but in case the 'constant'
        ``pad_mode`` is used with nonzero ``pad_const``, the
        derivative is given by the Gradient with ``pad_const=0``.

        Parameters
        ----------
        point : `domain` element, optional
            The point to take the derivative in. Does not change the result
            since the operator is affine.
        """
        if self.pad_mode == 'constant' and self.pad_const != 0:
            return Gradient(self.domain, self.range, self.method,
                            pad_mode=self.pad_mode,
                            pad_const=0)
        else:
            return self

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint is given by the negative `Divergence` with corrections for
        the method and padding.

        The Divergence is constructed from a ``space`` as a product space
        operator ``space^n --> space``, hence we need to provide the domain of
        this operator.
        """
        if not self.is_linear:
            raise ValueError('operator with nonzero pad_const ({}) is not'
                             ' linear and has no adjoint'
                             ''.format(self.pad_const))

        return - Divergence(domain=self.range, range=self.domain,
                            method=_ADJ_METHOD[self.method],
                            pad_mode=_ADJ_PADDING[self.pad_mode],
                            pad_const=self.pad_const)


class Divergence(PointwiseTensorFieldOperator):

    """Divergence operator for `DiscreteLp` spaces.

    Calls helper function `finite_diff` for each component of the input
    product space vector. For the adjoint of the `Divergence` operator to
    match the negative `Gradient` operator implicit zero is assumed.
    """

    def __init__(self, domain=None, range=None, method='forward',
                 pad_mode='constant', pad_const=0):
        """Initialize a new instance.

        Zero padding is assumed for the adjoint of the `Divergence`
        operator to match the negative `Gradient` operator.

        Parameters
        ----------
        domain : power space of `DiscreteLp`, optional
            Space of elements which the operator acts on.
            This is required if ``range`` is not given.
        range : `DiscreteLp`, optional
            Space of elements to which the operator maps.
            This is required if ``domain`` is not given.
        method : {'central', 'forward', 'backward'}, optional
            Finite difference method to be used
        pad_mode : string, optional
            The padding mode to use outside the domain.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis.

            ``'order2'``: Extend with second order accuracy (ensures continuity
            of the second derivative). This requires at least 3 values along
            each axis.

        pad_const : float, optional
            For ``pad_mode == 'constant'``, ``f`` assumes
            ``pad_const`` for indices outside the domain of ``f``

        Examples
        --------
        Initialize a Divergence opeator:

        >>> ran = odl.uniform_discr([0, 0], [3, 5], (3, 5))
        >>> dom = odl.ProductSpace(ran, ran.ndim)  # 2-dimensional
        >>> div = Divergence(dom)
        >>> div.range == ran
        True
        >>> div2 = Divergence(range=ran)
        >>> div2.domain == dom
        True
        >>> div3 = Divergence(domain=dom, range=ran)
        >>> div3.domain == dom
        True
        >>> div3.range == ran
        True

        Call the operator:

        >>> data = np.array([[0., 1., 2., 3., 4.],
        ...                  [1., 2., 3., 4., 5.],
        ...                  [2., 3., 4., 5., 6.]])
        >>> f = div.domain.element([data, data])
        >>> div_f = div(f)
        >>> print(div_f)
        [[2.0, 2.0, 2.0, 2.0, -3.0],
         [2.0, 2.0, 2.0, 2.0, -4.0],
         [-1.0, -2.0, -3.0, -4.0, -12.0]]

        Verify adjoint:

        >>> g = div.range.element(data ** 2)
        >>> adj_div_g = div.adjoint(g)
        >>> g.inner(div_f) / f.inner(adj_div_g)
        1.0
        """
        if domain is None and range is None:
            raise ValueError('either `domain` or `range` must be specified')

        if domain is None:
            if not isinstance(range, DiscreteLp):
                raise TypeError('`range` {!r} is not a DiscreteLp instance'
                                ''.format(range))
            domain = ProductSpace(range, range.ndim)

        if range is None:
            if not isinstance(domain, ProductSpace):
                raise TypeError('`domain` {!r} is not a ProductSpace instance'
                                ''.format(domain))
            range = domain[0]

        linear = not (pad_mode == 'constant' and pad_const != 0)
        super().__init__(domain, range, linear=linear)

        self.method, method_in = str(method).lower(), method
        if method not in _SUPPORTED_DIFF_METHODS:
            raise ValueError('`method` {} not understood'
                             ''.format(method_in))

        self.pad_mode, pad_mode_in = str(pad_mode).lower(), pad_mode
        if pad_mode not in _SUPPORTED_PAD_MODES:
            raise ValueError('`pad_mode` {} not understood'
                             ''.format(pad_mode_in))

        self.pad_const = range.field.element(pad_const)

    def _call(self, x, out=None):
        """Calculate the divergence of ``x``."""
        if out is None:
            out = self.range.element()

        ndim = self.range.ndim
        dx = self.range.cell_sides

        out_arr = out.asarray()
        tmp = np.empty(out.shape, out.dtype, order=out.space.order)
        for axis in range(ndim):
            finite_diff(x[axis], axis=axis, dx=dx[axis], method=self.method,
                        pad_mode=self.pad_mode,
                        pad_const=self.pad_const,
                        out=tmp)
            if axis == 0:
                out_arr[:] = tmp
            else:
                out_arr += tmp

        # self assignment: no overhead in the case asarray is a view
        out[:] = out_arr
        return out

    def derivative(self, point=None):
        """Return the derivative operator.

        The Divergence is usually linear, but in case the 'constant'
        ``pad_mode`` is used with nonzero ``pad_const``, the
        derivative is given by the Divergence with ``pad_const=0``.

        Parameters
        ----------
        point : `domain` element, optional
            The point to take the derivative in. Does not change the result
            since the operator is affine.
        """
        if self.pad_mode == 'constant' and self.pad_const != 0:
            return Divergence(self.domain, self.range, self.method,
                              pad_mode=self.pad_mode,
                              pad_const=0)
        else:
            return self

    @property
    def adjoint(self):
        """Adjoint of this operator.

        The adjoint is given by the negative `Gradient` with corrections for
        the method and padding.
        """
        if not self.is_linear:
            raise ValueError('operator with nonzero pad_const ({}) is not'
                             ' linear and has no adjoint'
                             ''.format(self.pad_const))

        return - Gradient(domain=self.range, range=self.domain,
                          method=_ADJ_METHOD[self.method],
                          pad_mode=_ADJ_PADDING[self.pad_mode])


class Laplacian(PointwiseTensorFieldOperator):

    """Spatial Laplacian operator for `DiscreteLp` spaces.

    Calls helper function `finite_diff` to calculate each component of the
    resulting product space vector.

    Outside the domain zero padding is assumed.
    """

    def __init__(self, space, pad_mode='constant', pad_const=0):
        """Initialize a new instance.

        Parameters
        ----------
        space : `DiscreteLp`
            Space of elements which the operator is acting on.
        pad_mode : string, optional
            The padding mode to use outside the domain.

            ``'constant'``: Fill with ``pad_const``.

            ``'symmetric'``: Reflect at the boundaries, not doubling the
            outmost values.

            ``'periodic'``: Fill in values from the other side, keeping
            the order.

            ``'order0'``: Extend constantly with the outmost values
            (ensures continuity).

            ``'order1'``: Extend with constant slope (ensures continuity of
            the first derivative). This requires at least 2 values along
            each axis.

            ``'order2'``: Extend with second order accuracy (ensures continuity
            of the second derivative). This requires at least 3 values along
            each axis where padding is applied.

        pad_const : float, optional
            For ``pad_mode == 'constant'``, ``f`` assumes
            ``pad_const`` for indices outside the domain of ``f``

        Examples
        --------
        >>> data = np.array([[ 0., 0., 0.],
        ...                  [ 0., 1., 0.],
        ...                  [ 0., 0., 0.]])
        >>> space = odl.uniform_discr([0, 0], [3, 3], [3, 3])
        >>> f = space.element(data)
        >>> lap = Laplacian(space)
        >>> print(lap(f))
        [[0.0, 1.0, 0.0],
         [1.0, -4.0, 1.0],
         [0.0, 1.0, 0.0]]
        """
        if not isinstance(space, DiscreteLp):
            raise TypeError('`space` {!r} is not a DiscreteLp instance'
                            ''.format(space))
        super().__init__(domain=space, range=space, linear=True)

        self.pad_mode, pad_mode_in = str(pad_mode).lower(), pad_mode
        if pad_mode not in _SUPPORTED_PAD_MODES:
            raise ValueError('`pad_mode` {} not understood'
                             ''.format(pad_mode_in))
        if pad_mode in ('order1', 'order1_adjoint',
                        'order2', 'order2_adjoint'):
            # TODO: Add these pad modes
            raise ValueError('`pad_mode` {} not implemented for Laplacian.'
                             ''.format(pad_mode_in))

        self.pad_const = space.field.element(pad_const)

    def _call(self, x, out=None):
        """Calculate the spatial Laplacian of ``x``."""
        if out is None:
            out = self.range.zero()
        else:
            out.set_zero()

        x_arr = x.asarray()
        out_arr = out.asarray()
        tmp = np.empty(out.shape, out.dtype, order=out.space.order)

        ndim = self.domain.ndim
        dx = self.domain.cell_sides

        for axis in range(ndim):
            # TODO: this can be optimized

            finite_diff(x_arr, axis=axis, dx=dx[axis] ** 2,
                        method='forward',
                        pad_mode=self.pad_mode,
                        pad_const=self.pad_const, out=tmp)

            out_arr[:] += tmp

            finite_diff(x_arr, axis=axis, dx=dx[axis] ** 2,
                        method='backward',
                        pad_mode=self.pad_mode,
                        pad_const=self.pad_const, out=tmp)

            out_arr[:] -= tmp

        out[:] = out_arr
        return out

    def derivative(self, point=None):
        """Return the derivative operator.

        The Laplacian is usually linear, but in case the 'constant'
        ``pad_mode`` is used with nonzero ``pad_const``, the
        derivative is given by the derivative with 0 ``pad_const``.

        Parameters
        ----------
        point : ``domain`` element
            The point to take the derivative in. Does not change the result
            since the operator is affine.
        """
        if self.pad_mode == 'constant' and self.pad_const != 0:
            return Laplacian(self.domain,
                             pad_mode=self.pad_mode,
                             pad_const=0)
        else:
            return self

    @property
    def adjoint(self):
        """Return the adjoint operator.

        The laplacian is self-adjoint, so this returns ``self``.
        """
        return self


def finite_diff(f, axis, dx=1.0, method='forward', out=None, **kwargs):
    """Calculate the partial derivative of ``f`` along a given ``axis``.

    In the interior of the domain of f, the partial derivative is computed
    using first-order accurate forward or backward difference or
    second-order accurate central differences.

    With padding the same method and thus accuracy is used on endpoints as
    in the interior i.e. forward and backward differences use first-order
    accuracy on edges while central differences use second-order accuracy at
    edges.

    Without padding one-sided forward or backward differences are used at
    the boundaries. The accuracy at the endpoints can then also be
    triggered by the edge order.

    The returned array has the same shape as the input array ``f``.

    Per default forward difference with dx=1 and no padding is used.

    Parameters
    ----------
    f : `array-like`
         An N-dimensional array.
    axis : int
        The axis along which the partial derivative is evaluated.
    dx : float, optional
        Scalar specifying the distance between sampling points along ``axis``.
    method : {'central', 'forward', 'backward'}, optional
        Finite difference method which is used in the interior of the domain
         of ``f``.
    out : `numpy.ndarray`, optional
         An N-dimensional array to which the output is written. Has to have
         the same shape as the input array ``f``.
    pad_mode : string, optional
        The padding mode to use outside the domain.

        ``'constant'``: Fill with ``pad_const``.

        ``'symmetric'``: Reflect at the boundaries, not doubling the
        outmost values.

        ``'periodic'``: Fill in values from the other side, keeping
        the order.

        ``'order0'``: Extend constantly with the outmost values
        (ensures continuity).

        ``'order1'``: Extend with constant slope (ensures continuity of
        the first derivative). This requires at least 2 values along
        each axis where padding is applied.

        ``'order2'``: Extend with second order accuracy (ensures continuity
        of the second derivative). This requires at least 3 values along
        each axis where padding is applied.

    pad_const : float, optional
        For ``pad_mode == 'constant'``, ``f`` assumes ``pad_const`` for
        indices outside the domain of ``f``

    Returns
    -------
    out : `numpy.ndarray`
        N-dimensional array of the same shape as ``f``. If ``out`` was
        provided, the returned object is a reference to it.

    Examples
    --------
    >>> f = np.array([ 0., 1., 2., 3., 4., 5., 6., 7., 8., 9.])

    >>> finite_diff(f, axis=0)
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  -9.])

    Without arguments the above defaults to:

    >>> finite_diff(f, axis=0, dx=1.0, method='forward', pad_mode='constant')
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  -9.])

    Parameters can be changed one by one:

    >>> finite_diff(f, axis=0, dx=0.5)
    array([ 2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  2.,  -18.])
    >>> finite_diff(f, axis=0, pad_mode='order1')
    array([ 1.,  1.,  1.,  1.,  1.,  1.,  1.,  1.,  1., 1.])

    Central differences and different edge orders:

    >>> finite_diff(0.5 * f**2, axis=0, method='central', pad_mode='order1')
    array([ 0.5,  1. ,  2. ,  3. ,  4. ,  5. ,  6. ,  7. ,  8. ,  8.5])
    >>> finite_diff(0.5 * f**2, axis=0, method='central', pad_mode='order2')
    array([-0.,  1.,  2.,  3.,  4.,  5.,  6.,  7.,  8.,  9.])

    In-place evaluation:

    >>> out = f.copy()
    >>> out is finite_diff(f, axis=0, out=out)
    True
    """
    f_arr = np.asarray(f)
    ndim = f_arr.ndim

    if f_arr.shape[axis] < 2:
        raise ValueError('in axis {}: at least two elements required, got {}'
                         ''.format(axis, f_arr.shape[axis]))

    if axis < 0:
        axis += ndim
    if not (0 <= axis < ndim):
        raise IndexError('`axis` {} outside the valid range 0 ... {}'
                         ''.format(axis, ndim - 1))

    dx, dx_in = float(dx), dx
    if dx <= 0 or not np.isfinite(dx):
        raise ValueError("`dx` must be positive, got {}".format(dx_in))

    method, method_in = str(method).lower(), method
    if method not in _SUPPORTED_DIFF_METHODS:
        raise ValueError('`method` {} was not understood'.format(method_in))

    pad_mode = kwargs.pop('pad_mode', 'constant')
    if pad_mode not in _SUPPORTED_PAD_MODES:
        raise ValueError('`pad_mode` {} not understood'
                         ''.format(pad_mode))
    pad_const = float(kwargs.pop('pad_const', 0))

    if out is None:
        out = np.empty_like(f_arr)
    else:
        if out.shape != f.shape:
            raise ValueError('expected output shape {}, got {}'
                             ''.format(f.shape, out.shape))

    if f_arr.shape[axis] < 2 and pad_mode == 'order1':
        raise ValueError("size of array to small to use 'order1', needs at "
                         "least 2 elements along axis {}.".format(axis))
    if f_arr.shape[axis] < 3 and pad_mode == 'order2':
        raise ValueError("size of array to small to use 'order2', needs at "
                         "least 3 elements along axis {}.".format(axis))

    if kwargs:
        raise ValueError('unkown keyword argument(s): {}'.format(kwargs))

    # create slice objects: initially all are [:, :, ..., :]

    # Swap axes so that the axis of interest is first. This is a O(1)
    # operation and is done to simplify the code below.
    out, out_in = np.swapaxes(out, 0, axis), out
    f_arr = np.swapaxes(f_arr, 0, axis)

    # Interior of the domain of f
    if method == 'central':
        # 1D equivalent: out[1:-1] = (f[2:] - f[:-2])/2.0
        np.subtract(f_arr[2:], f_arr[:-2], out=out[1:-1])
        out[1:-1] /= 2.0

    elif method == 'forward':
        # 1D equivalent: out[1:-1] = (f[2:] - f[1:-1])
        np.subtract(f_arr[2:], f_arr[1:-1], out=out[1:-1])

    elif method == 'backward':
        # 1D equivalent: out[1:-1] = (f[1:-1] - f[:-2])
        np.subtract(f_arr[1:-1], f_arr[:-2], out=out[1:-1])

    # Boundaries
    if pad_mode == 'constant':
        # Assume constant value c for indices outside the domain of ``f``

        # With padding the method used on endpoints is the same as in the
        # interior of the domain of f

        if method == 'central':
            out[0] = (f_arr[1] - pad_const) / 2.0
            out[-1] = (pad_const - f_arr[-2]) / 2.0

        elif method == 'forward':
            out[0] = f_arr[1] - f_arr[0]
            out[-1] = pad_const - f_arr[-1]

        elif method == 'backward':
            out[0] = f_arr[0] - pad_const
            out[-1] = f_arr[-1] - f_arr[-2]

    elif pad_mode == 'symmetric':
        # Values of f for indices outside the domain of f are replicates of
        # the edge values

        # With padding the method used on endpoints is the same as in the
        # interior of the domain of f

        if method == 'central':
            out[0] = (f_arr[1] - f_arr[0]) / 2.0
            out[-1] = (f_arr[-1] - f_arr[-2]) / 2.0

        elif method == 'forward':
            out[0] = f_arr[1] - f_arr[0]
            out[-1] = 0

        elif method == 'backward':
            out[0] = 0
            out[-1] = f_arr[-1] - f_arr[-2]

    elif pad_mode == 'symmetric_adjoint':
        # The adjoint case of symmetric

        if method == 'central':
            out[0] = (f_arr[1] + f_arr[0]) / 2.0
            out[-1] = (-f_arr[-1] - f_arr[-2]) / 2.0

        elif method == 'forward':
            out[0] = f_arr[1]
            out[-1] = -f_arr[-1]

        elif method == 'backward':
            out[0] = f_arr[0]
            out[-1] = -f_arr[-2]

    elif pad_mode == 'periodic':
        # Values of f for indices outside the domain of f are replicates of
        # the edge values on the other side

        if method == 'central':
            out[0] = (f_arr[1] - f_arr[-1]) / 2.0
            out[-1] = (f_arr[0] - f_arr[-2]) / 2.0

        elif method == 'forward':
            out[0] = f_arr[1] - f_arr[0]
            out[-1] = f_arr[0] - f_arr[-1]

        elif method == 'backward':
            out[0] = f_arr[0] - f_arr[-1]
            out[-1] = f_arr[-1] - f_arr[-2]

    elif pad_mode == 'order0':
        # Values of f for indices outside the domain of f are replicates of
        # the edge value.

        if method == 'central':
            out[0] = (f_arr[1] - f_arr[0]) / 2.0
            out[-1] = (f_arr[-1] - f_arr[-2]) / 2.0

        elif method == 'forward':
            out[0] = f_arr[1] - f_arr[0]
            out[-1] = 0

        elif method == 'backward':
            out[0] = 0
            out[-1] = f_arr[-1] - f_arr[-2]

    elif pad_mode == 'order0_adjoint':
        # Values of f for indices outside the domain of f are replicates of
        # the edge value.

        if method == 'central':
            out[0] = (f_arr[0] + f_arr[1]) / 2.0
            out[-1] = -(f_arr[-1] + f_arr[-2]) / 2.0

        elif method == 'forward':
            out[0] = f_arr[1]
            out[-1] = -f_arr[-1]

        elif method == 'backward':
            out[0] = f_arr[0]
            out[-1] = -f_arr[-2]

    elif pad_mode == 'order1':
        # Values of f for indices outside the domain of f are linearly
        # extrapolated from the inside.

        # independent of ``method``

        out[0] = f_arr[1] - f_arr[0]
        out[-1] = f_arr[-1] - f_arr[-2]

    elif pad_mode == 'order1_adjoint':
        # Values of f for indices outside the domain of f are linearly
        # extrapolated from the inside.

        if method == 'central':
            out[0] = f_arr[0] + f_arr[1] / 2.0
            out[-1] = -f_arr[-1] - f_arr[-2] / 2.0

            # Increment in case array is very short and we get aliasing
            out[1] -= f_arr[0] / 2.0
            out[-2] += f_arr[-1] / 2.0

        elif method == 'forward':
            out[0] = f_arr[0] + f_arr[1]
            out[-1] = -f_arr[-1]

            # Increment in case array is very short and we get aliasing
            out[1] -= f_arr[0]

        elif method == 'backward':
            out[0] = f_arr[0]
            out[-1] = -f_arr[-1] - f_arr[-2]

            # Increment in case array is very short and we get aliasing
            out[-2] += f_arr[-1]

    elif pad_mode == 'order2':
        # 2nd order edges

        out[0] = -(3.0 * f_arr[0] - 4.0 * f_arr[1] + f_arr[2]) / 2.0
        out[-1] = (3.0 * f_arr[-1] - 4.0 * f_arr[-2] + f_arr[-3]) / 2.0

    elif pad_mode == 'order2_adjoint':
        # Values of f for indices outside the domain of f are quadratically
        # extrapolated from the inside.

        if method == 'central':
            out[0] = 1.5 * f_arr[0] + 0.5 * f_arr[1]
            out[-1] = -1.5 * f_arr[-1] - 0.5 * f_arr[-2]

            # Increment in case array is very short and we get aliasing
            out[1] -= 1.5 * f_arr[0]
            out[2] += 0.5 * f_arr[0]
            out[-3] -= 0.5 * f_arr[-1]
            out[-2] += 1.5 * f_arr[-1]

        elif method == 'forward':
            out[0] = 1.5 * f_arr[0] + 1.0 * f_arr[1]
            out[-1] = -1.5 * f_arr[-1]

            # Increment in case array is very short and we get aliasing
            out[1] -= 2.0 * f_arr[0]
            out[2] += 0.5 * f_arr[0]
            out[-3] -= 0.5 * f_arr[-1]
            out[-2] += 1.0 * f_arr[-1]

        elif method == 'backward':
            out[0] = 1.5 * f_arr[0]
            out[-1] = -1.0 * f_arr[-2] - 1.5 * f_arr[-1]

            # Increment in case array is very short and we get aliasing
            out[1] -= 1.0 * f_arr[0]
            out[2] += 0.5 * f_arr[0]
            out[-3] -= 0.5 * f_arr[-1]
            out[-2] += 2.0 * f_arr[-1]
    else:
        raise NotImplementedError('unknown pad_mode')

    # divide by step size
    out /= dx

    return out_in


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()



# Copyright 2014-2016 The ODL development group
#
# This file is part of ODL.
#
# ODL is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# ODL is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with ODL.  If not, see <http://www.gnu.org/licenses/>.

"""Example functionals used in optimization."""

# Imports for common Python 2/3 codebase
from __future__ import print_function, division, absolute_import
from future import standard_library
standard_library.install_aliases()
from builtins import super

import numpy as np

from odl.solvers.functional.functional import Functional
from odl.operator.operator import Operator
from odl.space.base_ntuples import FnBase
from odl.space.npy_ntuples import MatVecOperator


__all__ = ('RosenbrockFunctional',)


class RosenbrockFunctional(Functional):

    """The well-known `Rosenbrock function`_ on ``R^n``.

    This function is usually used as a test problem in smooth optimization.

    Notes
    -----
    The functional is defined for :math:`x` in :math:`\\mathbb{R}^n`,
    :math:`n \\geq 2`, as

    .. math::
        \sum_{i=1}^{n - 1} c (x_{i+1} - x_i^2)^2 + (1 - x_i)^2

    Where :math:`c` is a constant usually set to 100 which determines how "ill-
    behaved" the function should be.
    It has a minimum at :math:`x = [1, \\dots, 1]`, independent of :math:`c`.

    There are two definitions of the n-dimensional Rosenbrock function found in
    the literature. One is the product of 2-dimensional Rosenbrock functions,
    which is not the one used here. This one extends the pattern of the 2d
    Rosenbrock function so all dimensions depend on each other in sequence.

    References
    ----------
    .. _Rosenbrock function: en.wikipedia.org/wiki/Rosenbrock_function
    """

    def __init__(self, space, scale=100.0):
        """Initialize a new instance.

        Parameters
        ----------
        space : `FnBase`
            Domain of the functional.
        scale : positive float, optional
            The scale ``c`` in the functional determining how "ill-behaved" the
            functional should be.

        Examples
        --------
        Initialize and call the functional:

        >>> r2 = odl.rn(2)
        >>> functional = RosenbrockFunctional(r2)
        >>> functional([1, 1])  # optimum is 0 at [1, 1]
        0.0
        >>> functional([0, 1])
        101.0

        The functional can also be used in higher dimensions:

        >>> r5 = odl.rn(5)
        >>> functional = RosenbrockFunctional(r5)
        >>> functional([1, 1, 1, 1, 1])
        0.0

        We can change how much the function is ill behaved via ``scale``:

        >>> r2 = odl.rn(2)
        >>> functional = RosenbrockFunctional(r2, scale=2)
        >>> functional([1, 1])  # optimum is still 0 at [1, 1]
        0.0
        >>> functional([0, 1])  # much lower variation
        3.0
        """
        self.scale = float(scale)
        if not isinstance(space, FnBase):
            raise ValueError('`space` must be an `FnBase`')
        if space.size < 2:
            raise ValueError('`space` must be at least two dimensional')
        super().__init__(space=space, linear=False, grad_lipschitz=np.inf)

    def _call(self, x):
        """Return ``self(x)``."""
        result = 0
        for i in range(0, self.domain.size - 1):
            result += (self.scale * (x[i + 1] - x[i] ** 2) ** 2 +
                       (x[i] - 1) ** 2)

        return result

    @property
    def gradient(self):
        """Gradient operator of the Rosenbrock functional."""
        functional = self
        c = self.scale

        class RosenbrockGradient(Operator):

            """The gradient operator of the Rosenbrock functional."""

            def __init__(self):
                """Initialize a new instance."""
                super().__init__(functional.domain, functional.domain,
                                 linear=False)

            def _call(self, x, out):
                """Apply the gradient operator to the given point."""
                for i in range(1, self.domain.size - 1):
                    out[i] = (2 * c * (x[i] - x[i - 1]**2) -
                              4 * c * (x[i + 1] - x[i]**2) * x[i] -
                              2 * (1 - x[i]))
                out[0] = (-4 * c * (x[1] - x[0] ** 2) * x[0] +
                          2 * (x[0] - 1))
                out[-1] = 2 * c * (x[-1] - x[-2] ** 2)

            def derivative(self, x):
                """The derivative of the gradient.

                This is also known as the Hessian.
                """

                # TODO: Implement optimized version of this that does not need
                # a matrix.
                shape = (functional.domain.size, functional.domain.size)
                matrix = np.zeros(shape)

                # Straightforward computation
                for i in range(0, self.domain.size - 1):
                    matrix[i, i] = (2 * c + 2 + 12 * c * x[i] ** 2 -
                                    4 * c * x[i + 1])
                    matrix[i + 1, i] = -4 * c * x[i]
                    matrix[i, i + 1] = -4 * c * x[i]
                matrix[-1, -1] = 2 * c
                matrix[0, 0] = 2 + 12 * c * x[0] ** 2 - 4 * c * x[1]
                return MatVecOperator(matrix, self.domain, self.range)

        return RosenbrockGradient()


if __name__ == '__main__':
    # pylint: disable=wrong-import-position
    from odl.util.testutils import run_doctests
    run_doctests()
